"""Concatenate Harvested Wood Products results from Roberto

Usage:

    ipython -i ~/eu_cbm/eu_cbm_hat/scripts/post_processing/hwp_load_from_excel.py

The following script moves HWP data from many Excel files to one csv file in
~/eu_cbm/eu_cbm_data/output_agg/hwp

"""

import pandas
from eu_cbm_hat import eu_cbm_data_pathlib
hwp_country_dir = eu_cbm_data_pathlib / "output_agg/hwp/countries"


#######################################
# Load and concatenate country tables #
#######################################
hwp = pandas.DataFrame()
for file_path in hwp_country_dir.iterdir():
    df = pandas.read_excel(file_path)
    hwp = pandas.concat([hwp, df])

hwp.reset_index(drop=True, inplace=True)


# Clean column names
cols = hwp.columns
selected_cols = cols[~cols.str.contains("named")]
selector = ~hwp["year"].isna()
hwp = hwp.loc[selector, selected_cols].copy()
hwp.rename(columns={"(All)": "all"}, inplace=True)
hwp.rename(columns=lambda x: str(x).lower(), inplace=True)

# Place scenario and country first
cols = ['scenario', 'country', 'year', 'hwp_sink_bau', 'hwp_sink_plus', 'hwp_sink_minor', 'crf', 'all']
hwp = hwp[cols].copy()

# Change column types
for col in ['hwp_sink_bau', 'hwp_sink_plus', 'hwp_sink_minor', 'crf', 'all']:
    hwp[col] = pandas.to_numeric(hwp[col], errors="coerce")

for col in ['scenario', 'country']:
    hwp[col] = hwp[col].astype(str)

hwp["year"] = hwp["year"].astype(int)

# Sort values
hwp.sort_values(["scenario", "country", "year"], inplace=True)

#############################################################
# Write to a CSV file to be archived in eu_cbm_data, in git #
#############################################################
# Note this file was first put in eu_cbm_data/domestic_harvest/hwp But this
# resulted in an error due to the make_combined function in
# ~/repos/eu_cbm/eu_cbm_hat/eu_cbm_hat/info/harvest.py
#
#     FileNotFoundError: [Errno 2] No such file or directory: <FilePath object
#     "/home/paul/repos/eu_cbm/eu_cbm_data/domestic_harvest/hwp/irw_harvest.csv">
#
# That is why it was moved to output_agg/hwp

file_path = eu_cbm_data_pathlib / "output_agg/hwp/hwp_2023.csv"
hwp.to_csv(file_path, index=False)




""" Run the EU_CBM_HAT scenario combinations

Usage:

    cd ~/repos/eu_cbm/eu_cbm_hat/scripts/running/
    # or on BDAP
    cd $HOME/eu_cbm/eu_cbm_hat/scripts/running/
    ipython -i run_scenario_combo.py -- --combo_name reference --last_year 2050
    ipython -i run_scenario_combo.py -- --combo_name pikssp2 --last_year 2070
    ipython -i run_scenario_combo.py -- --combo_name pikfair --last_year 2070
    ipython -i run_scenario_combo.py -- --combo_name pikssp2_fel1 --last_year 2070
    ipython -i run_scenario_combo.py -- --combo_name pikfair_fel1 --last_year 2070
    # Only run a few countries
    ipython -i run_scenario_combo.py -- --combo_name pikssp2 --last_year 2070 --countries LU CZ
    ipython -i run_scenario_combo.py -- --combo_name pikssp2 --last_year 2070 --countries IT

The version of p_umap in eu_cbm_hat/combos/base_combo.py Combination.__call__()
method was using a function that takes a list of runner objects as one its
argument. The problem was that `p_umap(run_country, runner_items, num_cpus=4)`
lead to `TypeError: cannot pickle 'sqlite3.Connection' object`. This is because
SQLite connections cannot be pickled and sent to subprocesses. Instead of
passing a runner as a function argument, the solution is to pass only the
country code as a function argument and have the function create the runner
object inside the sub-process directly.

Run, aggregate output and share on BDAP:

    cd $HOME/eu_cbm/eu_cbm_hat/scripts/running/
    ipython -i run_scenario_combo.py -- --combo_name reference --last_year 2050
    ipython -i run_scenario_combo.py -- --combo_name pikssp2_fel1 --last_year 2070
    ipython -i run_scenario_combo.py -- --combo_name pikfair_fel1 --last_year 2070
    cd $HOME/eu_cbm/eu_cbm_hat/scripts/post_processing
    ipython -i process_scenario_combo.py -- --combo_names reference pikssp2 pikfair
    rsync -zav $HOME/eu_cbm/eu_cbm_data/output_agg /eos/jeodpp/data/projects/SUSBIOM-TRADE/transfer/eu_cbm_data

"""

import argparse
from eu_cbm_hat.core.continent import continent

parser = argparse.ArgumentParser(description="Run the EU_CBM_HAT scenario combinations")
parser.add_argument(
    "--combo_name", type=str, help="Name of the scenario combo to be run"
)
parser.add_argument("--last_year", type=int, help="Last year to be simulated")
parser.add_argument(
    "--countries", nargs="+", default=None, help="List of country ISO2 codes"
)

shell_args = parser.parse_args()

LAST_YEAR = shell_args.last_year
COMBO_NAME = shell_args.combo_name
COUNTRIES = shell_args.countries

# Run the scenario combination for the given list of countries
continent.combos[COMBO_NAME].run(LAST_YEAR, COUNTRIES)



"""Process the stock output from the model"""
from typing import List, Union
from functools import cached_property
from eu_cbm_hat.post_processor.harvest import ton_carbon_to_m3_ub
from eu_cbm_hat.post_processor.convert import ton_carbon_to_m3_ob
import pandas as pd

class Stock:
    """Compute dw stock indicators

    Usage:

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['reference'].runners['LU'][-1]
        >>> runner.post_processor.stock.dw_stock_ratio("year")
        >>> runner.post_processor.stock.dw_stock_ratio(["year", "forest_type"])

    """

    def __init__(self, parent):
        self.parent = parent
        self.runner = parent.runner
        self.combo_name = self.runner.combo.short_name
        self.pools = self.parent.pools
        self.fluxes = self.parent.fluxes
        
    def volume_standing_stocks(self, groupby: Union[List[str], str] = None):
        """Estimate the mean ratio of standing stocks, ONLY merchantable"""
        if isinstance(groupby, str):
            groupby = [groupby]
        df = self.pools
        df = df.merge(self.parent.wood_density_bark_frac, on="forest_type")
        df = df[df ['status'] != 'NF']
        df["broad_standing_vol_ob"] = ton_carbon_to_m3_ob(df, "hardwood_merch")
        df["con_standing_vol_ob"] = ton_carbon_to_m3_ob(df, "softwood_merch")
        # Aggregate separately for softwood and hardwood
        #groupby = ['year','status', 'con_broad']
                  
        # Create a new column to identify the rows to be summed
        df_faws_ar = df[df['status'].isin(['AR', 'ForAWS'])]
        df_fnaws = df[df['status'].isin(['ForNAWS'])]
      
        # Group by the new column and the other columns
        df_faws = df_faws_ar.groupby(groupby).agg(
                    con_standing_for_volume=("con_standing_vol_ob", "sum"),
                    broad_standing_for_volume=("broad_standing_vol_ob", "sum"),
                    area = ("area", "sum")
                    ).reset_index()
            
        # rename the sum of AR and ForAWS as ForAWS
        df_faws['status'] = 'ForAWS'
        # Group by the new column and the other columns
        df_fnaws = df_fnaws.groupby(groupby).agg(
                    con_standing_for_volume=("con_standing_vol_ob", "sum"),
                    broad_standing_for_volume=("broad_standing_vol_ob", "sum"),
                    area = ("area", "sum")
                    ).reset_index()
        df_fnaws['status'] = 'ForNAWS'
        df_for = pd.concat([df_faws, df_fnaws])
        df_for['standing_stock_volume'] = df_for['con_standing_for_volume']+df_for['broad_standing_for_volume']
        df_for['standing_stock_volume_ha'] = df_for['standing_stock_volume']/df_for['area']
        df_for["combo_name"] = self.combo_name
        df_for["iso2_code"] = self.runner.country.iso2_code
        df_for["country"] = self.runner.country.country_name
        df_for=df_for.reset_index()
        return df_for
    
    def dw_stock_ratio(self, groupby: Union[List[str], str] = None):
        """Estimate the mean ratio of standing stocks, dead_wood to merchantable"""
        if isinstance(groupby, str):
            groupby = [groupby]
        df = self.pools

        # Aggregate separately for softwood and hardwood
        df_agg = df.groupby(groupby).agg(
            softwood_stem_snag_tc=("softwood_stem_snag", "sum"),
            softwood_merch_tc=("softwood_merch", "sum"),
            hardwood_stem_snag_tc=("hardwood_stem_snag", "sum"),
            hardwood_merch_tc=("hardwood_merch", "sum"),
            area=("area", sum),
            medium_tc=("medium_soil", "sum"),
        )

        df_agg.reset_index(inplace=True)
        df_agg["softwood_standing_dw_ratio"] = (
            df_agg["softwood_stem_snag_tc"] / df_agg["softwood_merch_tc"]
        )
        df_agg["hardwood_standing_dw_ratio"] = (
            df_agg["hardwood_stem_snag_tc"] / df_agg["hardwood_merch_tc"]
        )
        # agregate over con and broad
        df_agg["standing_dw_c_per_ha"] = (
            df_agg["hardwood_stem_snag_tc"] + df_agg["softwood_stem_snag_tc"]
        ) / df_agg["area"]
        df_agg["laying_dw_c_per_ha"] = df_agg["medium_tc"] / df_agg["area"]
        return df_agg

    def dw_contribution_harvest(self, groupby: Union[List[str], str] = None):
        """Estimate the mean ratio of standing stocks, dead_wood to merchantable"""
        if isinstance(groupby, str):
            groupby = [groupby]
        df = self.fluxes
        # Aggregate separately for softwood and hardwood
        df_agg = df.groupby("year").agg(
            softwood_merch_prod=("softwood_merch_to_product", "sum"),
            softwood_snag_prod=("softwood_stem_snag_to_product", "sum"),
            hardwood_merch_prod=("softwood_merch_to_product", "sum"),
            hardwood_snag_prod=("hardwood_stem_snag_to_product", "sum"),
        )
        df_agg["softwood_snag_harv_contrib"] = df_agg["softwood_snag_prod"] / (
            df_agg["softwood_snag_prod"] + df_agg["softwood_merch_prod"]
        )
        df_agg["hardwood_snag_harv_contrib"] = df_agg["hardwood_snag_prod"] / (
            df_agg["hardwood_snag_prod"] + df_agg["hardwood_merch_prod"]
        )
        return df_agg

    def df_agg(self, groupby: Union[List[str], str] = None):
        """Aggregated stock data

        Particularly interesting to get information the Dead Organic Matter stock.

        DOM stock distinguished by areas that have been harvested and areas not
        harvested:

            >>> from eu_cbm_hat.core.continent import continent
            >>> runner = continent.combos['reference'].runners['LU'][-1]
            >>> runner.post_processor.stock.df_agg(["year", "last_disturbance"])

        Check last disturbance type is 1 for rows where the condition is True

            >>> selected_cols = ["timestep", "disturbance_type", "last_disturbance_type", "time_since_last_disturbance"]
            >>> runner.post_processor.pools.query("last_disturbance_type == disturbance_type")[selected_cols]
            >>> #

        """
        if isinstance(groupby, str):
            groupby = [groupby]
        column_dict = {
            "merch": ["softwood_merch", "hardwood_merch"],
            "other": ["softwood_other", "hardwood_other"],
            "roots": [
                "softwood_fine_roots",
                "hardwood_fine_roots",
                "softwood_coarse_roots",
                "hardwood_coarse_roots",
            ],
            "foliage": ["softwood_foliage", "hardwood_foliage"],
            "litter": [
                "above_ground_very_fast_soil",
                "above_ground_fast_soil",
                "above_ground_slow_soil",
            ],
            "dead_wood": [
                "softwood_stem_snag",
                "softwood_branch_snag",
                "hardwood_branch_snag",
                "hardwood_stem_snag",
                "below_ground_fast_soil",
                "medium_soil",
            ],
        }
        df = self.parent.pools.copy()
        for key, cols in column_dict.items():
            df[key] = df[cols].sum(axis=1)
        df["dom"] = df["litter"] + df["dead_wood"]
        dist_types = self.parent.harvest.disturbance_types.copy()
        dist_types.rename(
            columns={
                "disturbance_type": "last_disturbance_type",
                "disturbance": "last_disturbance",
            },
            inplace=True,
        )
        df = df.merge(dist_types, on="last_disturbance_type")
        allowed_index = self.parent.classifiers_list
        allowed_index += [
            "year",
            "disturbance_type",
            "last_disturbance_type",
            "last_disturbance",
        ]
        missing_columns = set(groupby) - set(allowed_index)
        if bool(missing_columns):
            raise ValueError(
                f"Columns {missing_columns} are not allowed as groupby variables."
            )
        cols = ["area"] + list(column_dict.keys()) + ["dom"]
        df_agg = df.groupby(groupby)[cols].agg("sum").reset_index()
        return df_agg

    def dw_merch_stock_age_class(self, groupby: Union[List[str], str] = None):
        """Estimate the mean ratio of standing stocks, dead_wood to merchantable
        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['reference'].runners['LU'][-1]
        >>> runner.post_processor.stock.dw_merch_stock_age_class("year")
        """
        if isinstance(groupby, str):
            groupby = [groupby]
        df = self.pools
        df['age_class'] = df.age // 10 + 1
        df['age_class'] = 'AGEID' + df.age_class.astype(str)
        # Aggregate separately for softwood and hardwood
        df_agg = df.groupby(groupby+ ['age_class']).agg(
            softwood_stem_snag_tc=("softwood_stem_snag", "sum"),
            softwood_merch_tc=("softwood_merch", "sum"),
            hardwood_stem_snag_tc=("hardwood_stem_snag", "sum"),
            hardwood_merch_tc=("hardwood_merch", "sum"),
            #area=("area", sum),
            medium_tc=("medium_soil", "sum"),
        )
        df_agg.reset_index(inplace=True)
        df_agg["softwood_standing_dw_ratio"] = (
            df_agg["softwood_stem_snag_tc"] / df_agg["softwood_merch_tc"]
        )
        df_agg["hardwood_standing_dw_ratio"] = (
            df_agg["hardwood_stem_snag_tc"] / df_agg["hardwood_merch_tc"]
        )
        # agregate over con and broad
        #df_agg["standing_dw_c_per_ha"] = (
        #    df_agg["hardwood_stem_snag_tc"] + df_agg["softwood_stem_snag_tc"]
        #) / df_agg["area"]
        #df_agg["laying_dw_c_per_ha"] = df_agg["medium_tc"] / df_agg["area"]
        return df_agg



"""Get statistics on the country data for publication purposes

Country data is accessible through runner.orig_data

Note: cannot attach this method to the runner.post_processor, because
initializing runner.post processor requires data in the output directory. A
method attached to the post_processor would not run for countries which have no
model run output yet. Sometimes you want statistics on all countries for the
country orig data on a laptop where the model has not been run in all countries
yet i.e. a laptop where the output directory is not available for most
countries.

Print unique classifier values in all countries:

    >>> from eu_cbm_hat.post_processor.orig_data_stats import classifiers_all_countries
    >>> classifiers_all = classifiers_all_countries(combo_name="reference")
    >>> for this_cl in classifiers_all["classifier"].unique():
    >>>     selector = classifiers_all["classifier"] == this_cl
    >>>     if this_cl == "forest_type":
    >>>         # Remove NF forest types
    >>>         selector &= ~classifiers_all["classifier_value_id"].str.contains("NF")
    >>>     print("\n\n", this_cl)
    >>>     df = classifiers_all.loc[selector]
    >>>     unique_values = df["classifier_value_id"].unique()
    >>>     print(len(unique_values), unique_values)

Prepare a table of unique classifier values and description in all EU countries:

    >>> selector = classifiers_all["country"] != "ZZ"
    >>> classif_unique = classifiers_all.loc[selector].value_counts(["classifier", "classifier_value_id", "name"], sort=False)
    >>> classif_unique.to_csv("/tmp/classifiers.csv")

"""
from eu_cbm_hat.core.continent import continent
from functools import cached_property
from eu_cbm_hat.post_processor.agg_combos import apply_to_all_countries
from eu_cbm_hat.post_processor.agg_combos import place_combo_name_and_country_first


def classifiers_one_country(combo_name: str, iso2_code: str):
    """Classifiers of the country orig_data in one country"""
    runner = continent.combos[combo_name].runners[iso2_code][-1]
    df = runner.country.orig_data["classifiers"].copy()
    selector = df["classifier_value_id"] == "_CLASSIFIER"
    df.loc[selector, "classifier"] = df.loc[selector, "name"]
    df["classifier"].ffill(inplace=True)
    df = place_combo_name_and_country_first(df, runner)
    selector = df["classifier_value_id"] != '_CLASSIFIER'
    return df.loc[selector].copy()


def classifiers_all_countries(combo_name):
    """Classifiers of the country orig data in all countries"""
    df_all = apply_to_all_countries(classifiers_one_country, combo_name=combo_name)
    return df_all



"""before the update the AIDB from libcbm version 1 to libcbm version 2, 
check if there are mismatches between genus and forest typesin species table
"""

import pathlib
import shutil
import pandas
from cbm_defaults.update import db_updater
from eu_cbm_hat import eu_cbm_aidb_dir


from plumbing.databases.sqlite_database import SQLiteDatabase


countries_dir = pathlib.Path(eu_cbm_aidb_dir) / "countries"


def display_mismatch_forest_type_genus_type(db_path, table, columns):
    """
    >>> import sys
    >>> import pathlib
    >>> sys.path.append(str(pathlib.Path.home() /  "eu_cbm/eu_cbm_hat/scripts/conversion/libcbmv2/"))
    >>> from check_before_update_aidb_1x_to_2x import display_mismatch_forest_type_genus_type
    >>> from check_before_update_aidb_1x_to_2x import countries_dir
    >>> display_mismatch_forest_type_genus_type(countries_dir /"FI/aidb.db", table = "species", columns = ['genus_id','forest_type_id'])
    """
    db = SQLiteDatabase(str(db_path))
    df = db.read_df(table)
    
    # Group by genus_id and check if forest_type_id is unique
    unique_forest_types = df.groupby('genus_id')['forest_type_id'].nunique() > 1
    
    # Filter the data to show only the subsets where forest_type_id is not unique
    result = df[df['genus_id'].isin(unique_forest_types[unique_forest_types].index)]
    
    # Print the result
    return(result)


def display_duplicates_in_v1(db_path, table, columns):
    """
    Display duplications in a table based on specified columns.

    Args:
        db_path (pathlib.Path): The path to the database file.
        table (str): The name of the table to check for duplications.
        columns (list): A list of column names to consider for duplications.

    Returns:
        pd.DataFrame: A DataFrame containing the duplicate rows.
    """

    # Connect to the database and read the table into a DataFrame
    db = SQLiteDatabase(str(db_path))
    df = db.read_df(table)

    # Check if the columns exist in the DataFrame
    if not all(col in df.columns for col in columns):
        raise ValueError(f"One or more columns {columns} do not exist in the DataFrame.")

    # Find duplicates
    duplicates = df[df.duplicated(subset=columns, keep=False)]

    # Print the result
    print(f"Duplications in {table} based on {columns}:")
    print(duplicates)

    return duplicates


if __name__ == "__main__":
    aidbs = countries_dir.glob("**/aidb.db")
    for this_db in aidbs:
        # this_db = countries_dir / "ZZ" / "aidb.db"
        try:
            convert_aidb_to_v2(this_db)
        # Catch all errors. This can be restricted to fewer errors.
        # Generally observed errors are:
        # - sqlite3.OperationalError table disturbance_matrix_value has no
        #   column named index
        # - sqlite3.IntegrityError UNIQUE constraint failed:
        #   vol_to_bio_factor.id
        except Exception as e:
            print(f"Error {this_db}")
            print(e)
            print("\n\n")



"""
Aggregate scenario combination output and store them in the `eu_cbm_data/output_agg` directory.

For example, aggregate the output of the "reference" scenario combination and
save all post processing output tables for all countries to parquet files.

    >>> from eu_cbm_hat.post_processor.agg_combos import save_agg_combo_output
    >>> save_agg_combo_output("reference")

Other examples below explain how to run only some of the post processing steps.
This was useful as we were developing these output aggregation methods, we
often wanted to update only one type of output table for all scenarios at once,
and not have to wait to update all output tables for all scenarios.

- Save a specific data frame for all countries and all scenario combinations to
  parquet files

    >>> from eu_cbm_hat.post_processor.agg_combos import apply_to_all_combos
    >>> from eu_cbm_hat.post_processor.agg_combos import apply_to_all_countries
    >>> from eu_cbm_hat.post_processor.agg_combos import get_df_one_country
    >>> from eu_cbm_hat.post_processor.agg_combos import harvest_exp_prov_one_country
    >>> from eu_cbm_hat.post_processor.agg_combos import nai_one_country
    >>> from eu_cbm_hat.post_processor.agg_combos import output_agg_dir
    >>> from eu_cbm_hat.post_processor.agg_combos import pools_length_one_country

    >>> combos = ["pikssp2_fel1", "pikssp2_owc_max", "pikssp2_owc_min",
    >>>           "pikfair_fel1", "pikfair_owc_max", "pikfair_owc_min"]
    >>> apply_to_all_combos(pools_length_one_country, combos, "pools_length.parquet")
    >>> apply_to_all_combos(nai_one_country, combos, "nai_by_year_st.parquet", groupby=["status"])
    >>> apply_to_all_combos(harvest_exp_prov_one_country, combos,
    ...                     "hexprov_by_year.parquet", groupby=["year"])
    >>> apply_to_all_combos(area_one_country, combos, "area_st_ft_agecl.parquet",
    ...                     groupby=["year", "status", "forest_type", "age_class"])
    >>> apply_to_all_combos(harvest_exp_prov_one_country, combos, "hexprov_by_year_ft_dist.parquet",
    ...                     groupby=["year", "forest_type", "con_broad", "disturbance_type"])
    >>> apply_to_all_combos(get_df_one_country, combos, "stock_agg_by_year_ldist.parquet",
    ...                     groupby=["year", "last_disturbance"], runner_method_name="post_processor.stock.df_agg")

- Open the resulting parquet files to check the content of the data frames

    >>> from eu_cbm_hat.post_processor.agg_combos import read_agg_combo_output
    >>> sink = read_agg_combo_output(["reference", "pikfair"], "sink_by_year.parquet")
    >>> nai_st = read_agg_combo_output(combos, "nai_by_year_st_test_to_delete.parquet")
    >>> pools_length = read_agg_combo_output(combos, "pools_length.parquet")

- Save a specific data frame for all countries and only one scenario
  combination to parquet files

    >>> combo_name = "reference"
    >>> combo_dir = output_agg_dir / combo_name
    >>> sink = sink_all_countries(combo_name, "year")
    >>> sink.to_parquet(combo_dir / "sink_by_year_test_to_delete.parquet")
    >>> nai_st = apply_to_all_countries(nai_one_country, combo_name=combo_name, groupby=["status"])
    >>> nai_st.to_parquet(combo_dir / "nai_by_year_st_test_to_delete.parquet")
    >>> pools_length = apply_to_all_countries(pools_length_one_country, combo_name)
    >>> pools_length.to_parquet(combo_dir / "pools_length.parquet")

- Get data frames in all countries:

    >>> from eu_cbm_hat.post_processor.agg_combos import get_df_all_countries
    >>> # Events templates raw in all countries. This includes all scenarios,
    >>> # use with caution, it requires further filtering on the scenario column
    >>> # which is different than the combo_name column. The scenario column is given
    >>> # by runner.combo  runner.combo.config["events_templates"] variable.
    >>> events_templates_all = get_df_all_countries(
    >>>     combo_name="reference",
    >>>     runner_method_name="silv.events.raw"
    >>> )
    >>> # Load wood density and bark fraction in all countries.
    >>> wood_density_bark_all = get_df_all_countries(
    >>>     combo_name="reference",
    >>>     runner_method_name="post_processor.wood_density_bark_frac"
    >>> )
    >>> # Load inventory in all countries and sum the area by management type
    >>> inventory_all = get_df_all_countries(
    >>>     combo_name="reference",
    >>>     runner_method_name="country.orig_data.__getitem__",
    >>>     item=("mgmt", "inventory")
    >>> )
    >>> inventory_all["area"] = inventory_all["area"].astype(float)
    >>> inv_agg = inventory_all.groupby(["mgmt_strategy"]).agg(area = ("area","sum"))
    >>> inv_agg = inv_agg.assign(share = lambda x: x.area / x.area.sum())

- *Implementation note*: this script cannot be made a method of the
  combos/base_combo.py/Combination class because of circular references such as
  post_processor/harvest.py importing "continent" and "combined".

    >>> from eu_cbm_hat.info.harvest import combined
    >>> from eu_cbm_hat.core.continent import continent

    - To avoid these imports, functions in post_processor/harvest.py could be refactored.
    - Removing the "continent" could be done by changing functions to pass runner
      objects as arguments instead of creating the runner from the continent object.
    - The call to combined could be removed by loading the harvest demand table
      directly from CSV files.

"""

from typing import Union, List
import pandas
import warnings
from tqdm import tqdm
from p_tqdm import p_umap
from eu_cbm_hat.core.continent import continent
from eu_cbm_hat.post_processor.convert import ton_carbon_to_m3_ob
from eu_cbm_hat import eu_cbm_data_pathlib

# Define where to store the data
output_agg_dir = eu_cbm_data_pathlib / "output_agg"
output_agg_dir.mkdir(exist_ok=True)


def place_combo_name_and_country_first(df, runner):
    """Add combo name and country code to a data frame,
    place them as first columns"""
    df["combo_name"] = runner.combo.short_name
    df["iso2_code"] = runner.country.iso2_code
    df["country"] = runner.country.country_name
    cols = list(df.columns)
    cols = cols[-3:] + cols[:-3]
    return df[cols]


def get_df_one_country(combo_name, iso2_code, runner_method_name, **kwargs):
    """A generic function that returns the data frame output of the given
    runner method. See example use in the `get_df_all_countries()` function.
    """
    runner = continent.combos[combo_name].runners[iso2_code][-1]
    method = runner
    for method_name in runner_method_name.split("."):
        method = getattr(method, method_name)
    # If it's a @cached_property that returns a data frame directly
    if method.__class__.__name__ ==  "DataFrame":
        df = method
    # Otherwise call the method with additional arguments
    else:
        df = method(**kwargs)
    df = place_combo_name_and_country_first(df, runner)
    return df


def apply_to_all_countries(data_func, combo_name, **kwargs):
    """Apply a function to many countries"""
    df_all = pandas.DataFrame()
    country_codes = continent.combos[combo_name].runners.keys()
    for key in tqdm(country_codes):
        try:
            df = data_func(combo_name, key, **kwargs)
            df_all = pandas.concat([df, df_all])
        except FileNotFoundError as e_file:
            print(e_file)
        except ValueError as e_value:
            print(key, e_value)
    df_all.reset_index(inplace=True, drop=True)
    return df_all


def get_df_all_countries(combo_name, runner_method_name, **kwargs):
    """Get a data frame for all countries.

    Check wood density and bark fraction in all countries:

        >>> from eu_cbm_hat.post_processor.agg_combos import get_df_all_countries
        >>> wood_density_bark_all = get_df_all_countries(
        ...     combo_name="reference",
        ...     runner_method_name="post_processor.wood_density_bark_frac"
        ... )

    Check aggregated stock in all countries:

        >>> stock_agg_all = get_df_all_countries(
        ...     combo_name="reference",
        ...     runner_method_name="post_processor.stock.df_agg",
        ...     groupby = ["year", "last_disturbance"],
        ... )

    Area in all countries grouped by status, forest type, age and disturbances:

        >>> area_sfad_all = get_df_all_countries(
        ...     combo_name="reference",
        ...     runner_method_name="post_processor.area.df_agg",
        ...     groupby = ["year", 'status', "forest_type", "age","disturbance_type"],
        ... )

    Note: data types should be harmonized to avoid this error when writing to a parquet file
    ArrowTypeError: ("Expected bytes, got a 'int' object", 'Conversion failed for column climate with type object')
    """
    df_all = apply_to_all_countries(
        data_func=get_df_one_country,
        combo_name=combo_name,
        runner_method_name=runner_method_name,
        **kwargs,
    )
    return df_all

def apply_to_all_countries_and_save(args):
    """Get data for all countries and save it to a parquet file

    This function is to be used with
        - p_umap() in apply_to_all_combos().
        - and in save_agg_combo_output
    """
    data_func, combo_name, file_path, groupby, runner_method_name = args
    # Have to check if defined or not because combo_name is passed in **kwargs
    # TODO: check if it's possible to pass groupby, runner_method_name and
    # combo_name in kwargs
    if groupby is None and runner_method_name is None:
        print(f"Processing {combo_name} {data_func}.")
        df = apply_to_all_countries(data_func=data_func, combo_name=combo_name)
    elif runner_method_name is None:
        print(f"Processing {combo_name} {data_func} grouped by {groupby}.")
        df = apply_to_all_countries(
            data_func=data_func, combo_name=combo_name, groupby=groupby
        )
    else:
        msg = f"Processing {combo_name} {data_func} with {runner_method_name} "
        msg += f"grouped by {groupby}."
        print(msg)
        df = apply_to_all_countries(
            data_func=data_func, combo_name=combo_name, groupby=groupby, runner_method_name=runner_method_name
        )
    df.to_parquet(file_path)


def apply_to_all_combos(data_func, combo_names, file_name, groupby=None, runner_method_name=None):
    """Apply a function to all scenario combinations and save to parquet files

    This saves data for all countries in all scenario combinations into the
    given parquet file name. One file for each sub-directory in the
    eu_cbm_data/output_agg directory. These files can then be read and
    concatenated later with the read_agg_combo_output() function.

    Usage with the get_df_one_country function:

        >>> from eu_cbm_hat.post_processor.agg_combos import apply_to_all_combos
        >>> from eu_cbm_hat.post_processor.agg_combos import read_agg_combo_output
        >>> from eu_cbm_hat.post_processor.agg_combos import get_df_one_country
        >>> apply_to_all_combos(get_df_one_country, ["reference", "pikfair_fel1"], "stock_agg_example_to_delete.parquet", ["year", "last_disturbance"], "post_processor.stock.df_agg")
        >>> # Read the parquet file
        >>> stock_agg = read_agg_combo_output(["reference", "pikfair_fel1"], "stock_agg_example_to_delete.parquet")

   It can also be used with a custom function:
    
        >>> from eu_cbm_hat.post_processor.agg_combos import harvest_exp_prov_one_country
        >>> apply_to_all_combos(harvest_exp_prov_one_country, ["reference", "pikfair_fel1"], "hexprov_example_to_delete.parquet",
        ...                     groupby=["year", "forest_type", "con_broad", "disturbance_type"])
        >>> hexprov = read_agg_combo_output(["reference", "pikfair_fel1"], "hexprov_example_to_delete.parquet")

    """
    items = [
        (
            data_func,
            combo_name,
            output_agg_dir / combo_name / file_name,  # file path
            groupby,
            runner_method_name,
        )
        for combo_name in combo_names
    ]
    result = p_umap(apply_to_all_countries_and_save, items, num_cpus=4)
    return result


def save_agg_combo_output(combo_name: str):
    """Aggregate scenario combination output and store them in parquet files
    inside the `eu_cbm_data/output_agg` directory.

    Example save all post-processing data frames to `eu_cbm_data/output_agg`
    for one combo:

        >>> from eu_cbm_hat.post_processor.agg_combos import save_agg_combo_output
        >>> save_agg_combo_output("reference")

    Save for many scenario combinations in a loop:

        >>> for x in ["reference", "pikssp2", "pikfair"]:
        >>>     save_agg_combo_output(x)

    """
    combo_dir = output_agg_dir / combo_name
    combo_dir.mkdir(exist_ok=True)
    parameters = [
        {
            "data_func": harvest_exp_prov_one_country,
            "groupby": ["year"],
            "file_name": "hexprov_by_year.parquet",
        },
        {
            "data_func": harvest_exp_prov_one_country,
            "groupby": ["year", "forest_type", "con_broad", "disturbance_type"],
            "file_name": "hexprov_by_year_ft_dist.parquet",
        },
        {
            "data_func": sink_one_country,
            "groupby": "year",
            "file_name": "sink_by_year.parquet",
        },
        {
            "data_func": sink_one_country,
            "groupby": ["year", "status", "region"],
            "file_name": "sink_by_year_st_rg.parquet",
        },
        {
            "data_func": area_by_status_one_country,
            "groupby": None,
            "file_name": "area_by_year_status.parquet",
        },
        {
            "data_func": harvest_area_by_dist_one_country,
            "groupby": None,
            "file_name": "harvest_area_by_year_dist.parquet",
        },
        {
            "data_func": nai_one_country,
            "groupby": ["status"],
            "file_name": "nai_by_year_st.parquet",
        },
        {
            "data_func": area_by_age_class_one_country,
            "groupby": ["year", "status", "age_class"],
            "file_name": "area_by_age_class.parquet",
        },
        {
            "data_func": share_thinn_final_cut,
            "groupby": None,
            "file_name": "share_thinn_final_cut.parquet",
        },
        ]
    # List of parameters to be fed p_umap
    items = [
        (
            param["data_func"],
            combo_name,
            output_agg_dir / combo_name / param["file_name"],  # file path
            param["groupby"],
            param.get("runner_method_name", None)
        )
        for param in parameters
    ]
    result = p_umap(apply_to_all_countries_and_save, items, num_cpus=4)
    return result


#def save_agg_combo_output_legacy(combo_name: str):
#    """Aggregate scenario combination output and store them in parquet files
#    inside the `eu_cbm_data/output_agg` directory.
#    """
#    warnings.warn("This is the legacy version, use save_agg_combo_output instead")
#    combo_dir = output_agg_dir / combo_name
#    combo_dir.mkdir(exist_ok=True)
#    # Harvest expected provided by year
#    print(f"Processing {combo_name} harvest expected provided.")
#    hexprov_by_year = harvest_exp_prov_all_countries(combo_name, "year")
#    hexprov_by_year.to_parquet(combo_dir / "hexprov_by_year.parquet")
#    # Harvest expected provided by year, forest type and disturbance type
#    hexprov_by_year_ft_dist = harvest_exp_prov_all_countries(
#        combo_name, ["year", "forest_type", "disturbance_type"]
#    )
#    hexprov_by_year_ft_dist.to_parquet(combo_dir / "hexprov_by_year_ft_dist.parquet")
#    # Sink by year
#    print(f"Processing {combo_name} sink.")
#    sink = apply_to_all_countries(
#        sink_one_country, combo_name=combo_name, groupby="year"
#    )
#    sink.to_parquet(combo_dir / "sink_by_year.parquet")
#    # Sink by year and status
#    sink_ys = apply_to_all_countries(
#        sink_one_country, combo_name=combo_name, groupby=["year", "status"]
#    )
#    sink_ys.to_parquet(combo_dir / "sink_by_year_st.parquet")
#    print(f"Processing {combo_name} area.")
#    # Area by year and status
#    area_status = apply_to_all_countries(
#        area_by_status_one_country, combo_name=combo_name
#    )
#    area_status.to_parquet(combo_dir / "area_by_year_status.parquet")
#    print(f"Processing {combo_name} harvest area.")
#    harvest_area = apply_to_all_countries(
#        harvest_area_by_dist_one_country, combo_name=combo_name
#    )
#    harvest_area.to_parquet(combo_dir / "harvest_area_by_year_dist.parquet")
#    print(f"Processing {combo_name} Net Annual Increment.")
#    nai_sf = apply_to_all_countries(
#        nai_one_country, combo_name=combo_name, groupby=["status", "forest_type"]
#    )
#    nai_sf.to_parquet(combo_dir / "nai_by_year_st_ft.parquet")
#    nai_s = apply_to_all_countries(
#        nai_one_country, combo_name=combo_name, groupby=["status"]
#    )
#    nai_s.to_parquet(combo_dir / "nai_by_year_st.parquet")


def read_agg_combo_output(combo_name: list, file_name: str):
    """Read the aggregated combo output for the given list of combo names and
    the given file name. Return a concatenated data frame with data from all
    combos for that file.

    Example use:

        >>> from eu_cbm_hat.post_processor.agg_combos import read_agg_combo_output
        >>> sink = read_agg_combo_output(["reference", "pikfair"], "sink_by_year.parquet")
        >>> hexprov = read_agg_combo_output(["reference", "pikfair"], "hexprov_by_year.parquet")

    """
    df_all = pandas.DataFrame()
    df = pandas.DataFrame()
    for this_combo_name in combo_name:
        try:
            df = pandas.read_parquet(output_agg_dir / this_combo_name / file_name)
        except FileNotFoundError as error:
            print(error)
        df_all = pandas.concat([df_all, df])
    df_all.reset_index(inplace=True, drop=True)
    return df_all


def sink_one_country(
    combo_name: str,
    iso2_code: str,
    groupby: Union[List[str], str],
):
    """Sum the pools for the given country and add information on the combo
    country code

    The `groupby` argument specify the aggregation level. In addition to
    "year", one or more classifiers can be used for example "forest_type".

    The `pools_dict` argument is a dictionary mapping an aggregated pool name
    with the corresponding pools that should be aggregated into it. If you
    don't specify it, the function will used the default pools dict. The
    groupby argument makes it possible to specify how the sink rows will be
    grouped: by year, region, status and climate.

        >>> from eu_cbm_hat.post_processor.agg_combos import sink_one_country
        >>> ie_sink_y = sink_one_country("reference", "IE", groupby="year")
        >>> ie_sink_ys = sink_one_country("reference", "IE", groupby=["year", "status"])
        >>> lu_sink_y = sink_one_country("reference", "LU", groupby="year")
        >>> lu_sink_ys = sink_one_country("reference", "LU", groupby=["year", "status"])
        >>> lu_sink_yrc = sink_one_country("reference", "LU", groupby=["year", "region", "climate"])
        >>> hu_sink_y = sink_one_country("reference", "HU", groupby="year")

    Specify your own `pools_dict`:

        >>> pools_dict = {
        >>>     "living_biomass": [
        >>>         "softwood_merch",
        >>>         "softwood_other",
        >>>         "softwood_foliage",
        >>>         "softwood_coarse_roots",
        >>>         "softwood_fine_roots",
        >>>         "hardwood_merch",
        >>>         "hardwood_foliage",
        >>>         "hardwood_other",
        >>>         "hardwood_coarse_roots",
        >>>         "hardwood_fine_roots",
        >>>     ],
        >>>     "soil" : [
        >>>         "below_ground_very_fast_soil",
        >>>         "below_ground_slow_soil",
        >>>     ]
        >>> }
        >>> lu_sink_by_year = sink_one_country("reference", "LU", groupby="year", pools_dict=pools_dict)
        >>> index = ["year", "forest_type"]
        >>> lu_sink_by_y_ft = sink_one_country("reference", "LU", groupby=index, pools_dict=pools_dict)

    Sum flux pools and compute the sink

    Only return data for countries in which the model run was successful in
    storing the output data. Print an error message if the file is missing, but
    do not raise an error.

    Define a sink_all_countries function for the following examples

        >>> from eu_cbm_hat.post_processor.sink_all_countries import sink_one_country
        >>> def sink_all_countries(combo_name, groupby):
        >>>     df_all = apply_to_all_countries(
        >>>         sink_one_country, combo_name=combo_name, groupby=groupby
        >>>     )
        >>>     return df_all
        >>> sink = sink_all_countries("reference", "year")

    The purpose of this script is to compute the sink for all countries

    The following code summarises the flux_pool output for each country.

    For each year in each country:
    - aggregate the living biomass pools
    - compute the stock change
    - multiply by -44/12 to get the sink.


    Usage example (see also functions documentation bellow).

    Get the biomass sink for 2 scenarios:

        >>> import pandas
        >>> # Replace these by the relevant scenario combinations
        >>> sinkfair = sink_all_countries("pikfair", "year")
        >>> sinkbau =  sink_all_countries("pikssp2", "year")
        >>> df_all = pandas.concat([sinkfair, sinkbau])
        >>> df_all.reset_index(inplace=True, drop=True)
        >>> df_all.sort_values("country", inplace=True)

    Note the area is stable through time, transition rules only make it move from
    one set of classifiers to another set of classifiers.

        from eu_cbm_hat.core.continent import continent
        runner = continent.combos["pikfair"].runners["IE"][-1]
        classifiers = runner.output.classif_df
        index = ["identifier", "timestep"]
        pools = runner.output["pools"].merge(classifiers, "left", on=index)
        area_status = (pools.groupby(["timestep", "status"])["area"]
                       .agg("sum")
                       .reset_index()
                       .pivot(columns="status", index="timestep", values="area")
                       )
        cols = df.columns
        area_status["sum"] = area_status.sum(axis=1)

    The following code chunk is a justification of why we need to look at the
    carbon content of soils in this convoluted way. Because a few afforested plots
    have AR present in the first time step, then we cannot compute a difference to
    the previous time step, and we need . In Ireland for example the following
    identifiers have "AR" present in their first time step:

        from eu_cbm_hat.core.continent import continent
        runner = continent.combos['reference'].runners['IE'][-1]
        # Load pools
        classifiers = runner.output.classif_df
        classifiers["year"] = runner.country.timestep_to_year(classifiers["timestep"])
        index = ["identifier", "timestep"]
        df = runner.output["pools"].merge(classifiers, "left", on=index)
        # Show the first time step of each identifier with AR status
        df["min_timestep"] = df.groupby("identifier")["timestep"].transform(min)
        selector = df["status"].str.contains("AR")
        selector &= df["timestep"] == df["min_timestep"]
        ar_first = df.loc[selector]
        ar_first[["identifier", "timestep", "status", "area", "below_ground_slow_soil"]]

    Aggregate by year, status, region and climate

    TODO: complete this example
    Compute the sink along the status
    Provide an example that Aggregate columns that contains "AR", such as
    ["AR_ForAWS", "AR_ForNAWS"] to a new column called "AR_historical".

        >>> for new_column, columns_to_sum in aggregation_dict.items():
        >>>     df[new_column] = df[columns_to_sum].sum(axis=1)
        >>>     df.drop(columns=columns_to_sum, inplace=True)

    """
    if "year" not in groupby:
        raise ValueError("Year has to be in the group by variables")
    if isinstance(groupby, str):
        groupby = [groupby]
    runner = continent.combos[combo_name].runners[iso2_code][-1]
    df_agg = runner.post_processor.sink.df_agg(groupby=groupby)
    df_agg = place_combo_name_and_country_first(df_agg, runner)
    return df_agg


def area_one_country(combo_name: str, iso2_code: str, groupby: Union[List[str], str]):
    """Harvest provided in one country

    Usage:

        >>> from eu_cbm_hat.post_processor.agg_combos import area_one_country
        >>> area_one_country("reference", "ZZ", ["year", 'status', "forest_type", "age","disturbance_type"])

    """
    runner = continent.combos[combo_name].runners[iso2_code][-1]
    df_agg = runner.post_processor.area.df_agg(groupby)
    df_agg = place_combo_name_and_country_first(df_agg, runner)
    return df_agg




def area_by_status_one_country(combo_name: str, iso2_code: str):
    """Area in wide format with one column for each status.

    This table describes the movement from non forested to forested areas.
    Afforestation and deforestation influence the changes in area. Total area
    remains the same.

    Usage:

        >>> from eu_cbm_hat.post_processor.area import area_by_status_one_country
        >>> from eu_cbm_hat.post_processor.area import apply_to_all_countries
        >>> area_by_status_one_country("reference", "ZZ")
        >>> ast_ie = area_by_status_one_country("reference", "IE")
        >>> # Load data for all countries
        >>> ast = apply_to_all_countries(area_by_status_one_country, combo_name="reference")
        >>> # Place total area column last
        >>> cols = list(ast.columns)
        >>> cols.remove("total_area")
        >>> cols += ["total_area"]
        >>> ast = ast[cols]

    """
    groupby = ["year", "status", "disturbance_type"]
    df = area_one_country(combo_name=combo_name, iso2_code=iso2_code, groupby=groupby)
    # Change disturbance deforestation to status D
    selector = df["disturbance_type"] == 7
    df.loc[selector, "status"] = "D"
    # Aggregate
    index = ["year", "status"]
    df = df.groupby(index)["area"].agg("sum").reset_index()
    # Pivot to wide format
    df_wide = df.pivot(index="year", columns="status", values="area")
    # Add the total area
    df_wide["total_area"] = df_wide.sum(axis=1)
    df_wide.reset_index(inplace=True)
    # Remove the sometimes confusing axis name
    df_wide.rename_axis(columns=None, inplace=True)
    # Place combo name, country code as first columns
    df_wide["combo_name"] = combo_name
    df_wide["iso2_code"] = iso2_code
    cols = list(df_wide.columns)
    cols = cols[-2:] + cols[:-2]
    return df_wide[cols]

""
def area_by_age_class_one_country(combo_name: str, iso2_code: str, groupby: Union[List[str], str]):
    """Area in wide format with one column for ageclass.
        from eu_cbm_hat.post_processor.area import aarea_by_age_class_one_country
        area_by_age_class_one_country("reference", ["year", "status", "con_broad", "disturbance_type"])
    """
    groupby = ["year", "status", "age_class"]
    df = area_one_country(combo_name=combo_name, iso2_code=iso2_code, groupby=groupby)
    df=df[['combo_name', 'iso2_code', 'country', 'year', 'age_class', 'area']]
    return df

def area_by_age_class_all_countries(combo_name: str, groupby: Union[List[str], str]):
    """NAI area by status in wide format for all countries in the given scenario combination.

    >>> from eu_cbm_hat.post_processor.area import area_by_age_class_all_countries
    >>> area_by_age_class_all_countries("reference", ["year", "status", "con_broad", "disturbance_type"])

    """
    df_all = apply_to_all_countries(
        area_by_age_class_one_country, combo_name=combo_name, groupby=groupby
    )
    return df_all

""
def share_thinn_final_cut(combo_name: str, iso2_code: str):
    """Area in wide format with one column for each status.

    Usage:

        >>> from eu_cbm_hat.post_processor.agg_combos import share_thinn_final_cut
        >>> hare_thinn_final_cut("reference", "LU")

    """
    groupby = ['year', 'con_broad', 'silv_practice']
    runner = continent.combos[combo_name].runners[iso2_code][-1]
    df_shares = runner.post_processor.harvest.provided_shares
    df_shares ["scenario"] = runner.combo.short_name
    df_shares ["region"] = runner.country.country_name
    return df_shares

""
def harvest_area_by_dist_one_country(combo_name: str, iso2_code: str):
    """Area in wide format with one column for each status.

    Usage:

        >>> from eu_cbm_hat.post_processor.agg_combos import harvest_area_by_dist_one_country
        >>> harvest_area_by_dist_one_country("reference", "LU")

    """
    groupby = ["year", "disturbance_type", "disturbance"]
    runner = continent.combos[combo_name].runners[iso2_code][-1]
    df = runner.post_processor.harvest.area_agg(groupby=groupby)
    df = place_combo_name_and_country_first(df, runner)
    return df


def nai_one_country(combo_name: str, iso2_code: str, groupby: Union[List[str], str]):
    """Net Annual Increment data by status and forest type

    Usage:

        >>> from eu_cbm_hat.post_processor.agg_combos import nai_one_country
        >>> nai_one_country("reference", "LU", ["status"])
        >>> nai_one_country("reference", "LU", ["status", "forest_type"])

    """
    runner = continent.combos[combo_name].runners[iso2_code][-1]
    df = runner.post_processor.nai.df_agg(groupby=groupby)
    df = place_combo_name_and_country_first(df, runner)
    return df


def nai_all_countries(combo_name: str, groupby: Union[List[str], str]):
    """NAI area by status in wide format for all countries in the given scenario combination.

    >>> from eu_cbm_hat.post_processor.area import nai_all_countries
    >>> nai_all_countries("reference", ["year", "status", "con_broad", "disturbance_type"])

    """
    df_all = apply_to_all_countries(
        nai_one_country, combo_name=combo_name, groupby=groupby
    )
    return df_all

def nai_con_broad_one_country(combo_name: str, iso2_code: str, groupby: Union[List[str], str]):
    """Net Annual Increment data by status and con_broad

    Usage:

        >>> from eu_cbm_hat.post_processor.agg_combos import nai_one_country
        >>> nai_one_country("reference", "LU", ["status"])
        >>> nai_one_country("reference", "LU", ["status", "forest_type"])

    """
    runner = continent.combos[combo_name].runners[iso2_code][-1]
    df = runner.post_processor.nai.df_agg_con_broad(groupby=groupby)
    df = place_combo_name_and_country_first(df, runner)
    return df

def nai_con_broad_all_countries(combo_name: str, groupby: Union[List[str], str]):
    """NAI area by status in wide format for all countries in the given scenario combination.

    >>> from eu_cbm_hat.post_processor.area import nai_all_countries
    >>> nai_all_countries("reference", ["year", "status", "con_broad", "disturbance_type"])

    """
    df_all = apply_to_all_countries(
        nai_con_broad_one_country, combo_name=combo_name, groupby=groupby
    )
    cols_to_keep = ['combo_name', 'country', 'year', 'status', 'con_broad',
       'area', 'nai_merch', 'gai_merch','nai_agb', 'gai_agb', 'nai_merch_ha', 'gai_merch_ha', 'nai_agb_ha','gai_agb_ha']
    df_all = df_all[cols_to_keep]
    return df_all

def weighted_nai_con_broad_one_country(combo_name: str, iso2_code: str, groupby: Union[List[str], str]):
    """Net Annual Increment data weighted by status

    Usage:

        >>> from eu_cbm_hat.post_processor.agg_combos import nai_one_country
        >>> nai_one_country("reference", "LU", ["status"])
        >>> nai_one_country("reference", "LU", ["status", "forest_type"])

    """
    runner = continent.combos[combo_name].runners[iso2_code][-1]
    df = runner.post_processor.nai.df_agg_con_broad(groupby=groupby)
    df = place_combo_name_and_country_first(df, runner)
    # Calculate the weighted values for nai_merch_ha and gai_merch_ha by multiplying by area
    df['weighted_nai'] = df['nai_merch_ha'] * df['area']
    df['weighted_gai'] = df['gai_merch_ha'] * df['area']
    
    # Group by the necessary columns and sum the weighted values and the area
    grouped = df.groupby(['country', 'year', 'con_broad']).agg(
        total_weighted_nai=('weighted_nai', 'sum'),
        total_weighted_gai=('weighted_gai', 'sum'),
        total_area=('area', 'sum')
    ).reset_index()
    
    # Calculate the weighted averages
    grouped['avg_nai_merch_ha'] = grouped['total_weighted_nai'] / grouped['total_area']
    grouped['avg_gai_merch_ha'] = grouped['total_weighted_gai'] / grouped['total_area']
    
    # Select the relevant columns for the final output
    final_df = grouped[['country', 'year', 'con_broad', 'avg_nai_merch_ha', 'avg_gai_merch_ha']]
    return final_df


def weighted_nai_con_broad_all_countries(combo_name: str, groupby: Union[List[str], str]):
    """NAI area by status in wide format for all countries in the given scenario combination.

    >>> from eu_cbm_hat.post_processor.area import nai_all_countries
    >>> nai_all_countries("reference", ["year", "status", "con_broad", "disturbance_type"])

    """
    df_all = apply_to_all_countries(
        weighted_nai_con_broad_one_country, combo_name=combo_name, groupby=groupby
    )
    return df_all

def area_all_countries(combo_name: str, groupby: Union[List[str], str]):
    """Harvest area by status in wide format for all countries in the given scenario combination.

    >>> from eu_cbm_hat.post_processor.area import area_all_countries
    >>> area_all_countries("reference", ["year", "status", "con_broad", "disturbance_type"])

    """
    df_all = apply_to_all_countries(
        area_one_country, combo_name=combo_name, groupby=groupby
    )
    return df_all


def harvest_exp_prov_one_country(
    combo_name: str, iso2_code: str, groupby: Union[List[str], str]
):
    """Harvest excepted provided in one country

    There is a groupby  argument because we get the harvest expected from the
    hat output of disturbances allocated by hat which are allocated at some
    level of classifier groupings (other classifiers might have question marks
    i.e. where harvest can be allocated to any value of that particular
    classifier).

    In case the groupby argument is equal to "year", we also add the harvest
    demand from the economic model.

    Usage:

        >>> from eu_cbm_hat.post_processor.agg_combos import harvest_exp_prov_one_country
        >>> import pandas
        >>> pandas.set_option('display.precision', 0) # Display rounded numbers
        >>> harvest_exp_prov_one_country("reference", "ZZ", "year")
        >>> harvest_exp_prov_one_country("reference", "ZZ", ["year", "forest_type"])
        >>> harvest_exp_prov_one_country("reference", "ZZ", ["year", "disturbance_type"])

    """
    if isinstance(groupby, str):
        groupby = [groupby]

    # TODO: current version of harvest_exp_one_country() only contains HAT
    # disturbances. This should also provide static events that generate fluxes
    # to products especially in the historical period
    runner = continent.combos[combo_name].runners[iso2_code][-1]
    df = runner.post_processor.harvest.expected_provided(groupby=groupby)
    df = place_combo_name_and_country_first(df, runner)
    return df


def harvest_exp_prov_all_countries(combo_name: str, groupby: Union[List[str], str]):
    """Information on both harvest expected and provided for all countries in
    the combo_name.

    Some countries might have NA values. If the model didn't run successfully
    for those countries i.e. the output flux table was missing.

    Example use:

        >>> from eu_cbm_hat.post_processor.agg_combos import harvest_exp_prov_all_countries
        >>> harvest_exp_prov_all_countries("reference", "year")
        >>> harvest_exp_prov_all_countries("reference", ["year", "forest_type", "disturbance_type"])

    """
    df_all = apply_to_all_countries(
        harvest_exp_prov_one_country, combo_name=combo_name, groupby=groupby
    )
    return df_all

def volume_stock_one_country(
    combo_name: str, iso2_code: str, groupby: Union[List[str], str]
):
    """
    Usage:

        >>> from eu_cbm_hat.post_processor.agg_combos import volume_stock_one_country
        >>> import pandas
        >>> pandas.set_option('display.precision', 0) # Display rounded numbers
        >>> volume_stock_one_country("reference", "ZZ", "year")
        >>> volume_stock_one_country("reference", "ZZ", ["year", "status"])

    """
    if isinstance(groupby, str):
        groupby = [groupby]
    runner = continent.combos[combo_name].runners[iso2_code][-1]
    df = runner.post_processor.stock.volume_standing_stocks(groupby=groupby)
    print(df.head(3))
    df = place_combo_name_and_country_first(df, runner)
    return df

def volume_stock_all_countries(combo_name: str, groupby: Union[List[str], str]):
    """Information on both harvest expected and provided for all countries in
    the combo_name.
    Example use:

        >>> from eu_cbm_hat.post_processor.agg_combos import harvest_exp_prov_all_countries
        >>> harvest_exp_prov_all_countries("reference", "year")
        >>> harvest_exp_prov_all_countries("reference", ["year", "status"])

    """
    df_all = apply_to_all_countries(
        volume_stock_one_country, combo_name=combo_name, groupby=groupby
    )
    return df_all

def soc_one_country(combo_name: str, iso2_code: str, groupby: Union[List[str], str]):
    """Harvest provided in one country
    Usage:

        >>> from eu_cbm_hat.post_processor.agg_combos import soc_one_country
        >>> dw_zz = soc_one_country("reference", "ZZ", ["year", 'status', "disturbance_type"])

    """
    index = ["identifier", "timestep"]
    runner = continent.combos[combo_name].runners[iso2_code][-1]
    # Load Area
    cols_to_keep = [
        "area",
        "softwood_merch",
        "hardwood_merch",
        "medium_soil",
        "softwood_stem_snag",
        "hardwood_stem_snag",
        # new ones
        "above_ground_very_fast_soil",
        "below_ground_very_fast_soil",
        "above_ground_fast_soil",
        "below_ground_fast_soil",
        "above_ground_slow_soil",
        "below_ground_slow_soil",
        "softwood_branch_snag",
        "hardwood_branch_snag",
    ]
    df = runner.output["pools"][index + cols_to_keep]
    df["year"] = runner.country.timestep_to_year(df["timestep"])
    # Add classifiers
    df = df.merge(runner.output.classif_df, on=index)
    # Disturbance type information
    dist = runner.output["parameters"][index + ["disturbance_type"]]
    df = df.merge(dist, on=index)
    # Aggregate
    # df_agg = df.groupby(groupby)["medium_soil"].agg("sum").reset_index()

    # Aggregate separately for softwood and hardwood
    df_agg = df.groupby(groupby).agg(
        softwood_stem_snag_tc=("softwood_stem_snag", "sum"),
        softwood_merch_tc=("softwood_merch", "sum"),
        hardwood_stem_snag_tc=("hardwood_stem_snag", "sum"),
        hardwood_merch_tc=("hardwood_merch", "sum"),
        medium_tc=("medium_soil", "sum"),
        # new ones
        above_ground_very_fast_soil_tc=("above_ground_very_fast_soil", sum),
        below_ground_very_fast_soil_tc=("below_ground_very_fast_soil", sum),
        above_ground_fast_soil_tc=("above_ground_fast_soil", sum),
        below_ground_fast_soil_tc=("below_ground_fast_soil", sum),
        above_ground_slow_soil_tc=("above_ground_slow_soil", sum),
        below_ground_slow_soil_tc=("below_ground_slow_soil", sum),
        area=("area", "sum"),
    )
    df_agg.reset_index(inplace=True)
    df_agg["softwood_standing_dw_ratio"] = (
        df_agg["softwood_stem_snag_tc"] / df_agg["softwood_merch_tc"]
    )
    df_agg["hardwood_standing_dw_ratio"] = (
        df_agg["hardwood_stem_snag_tc"] / df_agg["hardwood_merch_tc"]
    )
    # agregate over con and broad
    df_agg["standing_dw_c_per_ha"] = (
        df_agg["hardwood_stem_snag_tc"] + df_agg["softwood_stem_snag_tc"]
    ) / df_agg["area"]
    df_agg["laying_dw_c_per_ha"] = df_agg["medium_tc"] / df_agg["area"]

    # Place combo name, country code and country name as first columns
    df_agg["combo_name"] = combo_name
    df_agg["iso2_code"] = runner.country.iso2_code
    df_agg["country"] = runner.country.country_name
    cols = list(df_agg.columns)
    # cols = ['softwood_stem_snag_tc', 'softwood_merch_tc',
    # 'hardwood_stem_snag_tc', 'hardwood_merch_tc', 'area', 'medium_tc',
    # 'softwood_standing_dw_ratio', 'hardwood_standing_dw_ratio',
    # 'standing_dw_c_per_ha', 'laying_dw_c_per_ha']
    cols = cols[-3:] + cols[:-3]
    return df_agg[cols]


def soc_all_countries(combo_name: str, groupby: Union[List[str], str]):
    """Harvest area by status in wide format for all countries in the given scenario combination."""
    df_all = apply_to_all_countries(
        soc_one_country, combo_name=combo_name, groupby=groupby
    )
    return df_all


def pools_length_one_country(
    combo_name: str,
    iso2_code: str,
):
    """Number of rows in the pools table

    Usage:

        >>> from eu_cbm_hat.post_processor.agg_combos import pools_length_one_country
        >>> pools_length_one_country("reference", "LU")

    """
    runner = continent.combos[combo_name].runners[iso2_code][-1]
    pools = runner.post_processor.pools
    df = pools.value_counts(["year"], sort=False).reset_index()
    df = place_combo_name_and_country_first(df, runner)
    # Fix for older versions of pandas
    df.rename(columns={0: "count"}, inplace=True)
    return df



"""
The purpose of this script is to compute the sink for one country
"""

from typing import List, Union
from functools import cached_property
import numpy as np
import pandas


POOLS_DICT = {
    "living_biomass": [
        "softwood_merch",
        "softwood_other",
        "softwood_foliage",
        "softwood_coarse_roots",
        "softwood_fine_roots",
        "hardwood_merch",
        "hardwood_foliage",
        "hardwood_other",
        "hardwood_coarse_roots",
        "hardwood_fine_roots",
    ],
    "litter": [
        "above_ground_very_fast_soil",
        "above_ground_fast_soil",
        "above_ground_slow_soil",
    ],
    "dead_wood": [
        "softwood_stem_snag",
        "softwood_branch_snag",
        "hardwood_branch_snag",
        "hardwood_stem_snag",
        "below_ground_fast_soil",
        "medium_soil",
    ],
    "soil": [
        "below_ground_very_fast_soil",
        "below_ground_slow_soil",
    ],
}

FLUXES_DICT = {
    "loss_from_living_biomass": [
        # transfers to products
        "softwood_merch_to_product",
        "softwood_other_to_product",
        "hardwood_merch_to_product",
        "hardwood_other_to_product",
        # any direct flux to air
        "disturbance_merch_to_air",
        "disturbance_fol_to_air",
        "disturbance_oth_to_air",
        "disturbance_coarse_to_air",
        "disturbance_fine_to_air",
    ],
    "loss_from_litter": [
        "decay_v_fast_ag_to_air",
        "decay_fast_ag_to_air",
        "decay_slow_ag_to_air",
    ],
    "loss_from_dead_wood": [
        "softwood_stem_snag_to_product",
        "softwood_branch_snag_to_product",
        "hardwood_stem_snag_to_product",
        "hardwood_branch_snag_to_product",
        "decay_sw_stem_snag_to_air",
        "decay_sw_branch_snag_to_air",
        "decay_hw_stem_snag_to_air",
        "decay_hw_branch_snag_to_air",
        "decay_fast_bg_to_air",
        "decay_medium_to_air",
    ],
    "loss_from_soil": [
        "decay_v_fast_bg_to_air",
        "decay_slow_bg_to_air",
    ],
    "loss_from_non_co2_emissions": [
        "disturbance_bio_ch4_emission",
        "disturbance_bio_co_emission",
        "disturbance_domch4_emission",
        "disturbance_domco_emission",
    ],
}


def generate_all_combinations_and_fill_na(df, groupby):
    """Generate a DataFrame with all combinations of year, status, region, and
    climate.
    """
    groupby_area_diff = groupby.copy()
    # Prepare all combinations of groupby variables except year
    groupby_area_diff.remove("year")
    all_groups = df[groupby_area_diff].drop_duplicates()
    years = list(df["year"].unique())
    combi_dict = {
        "year": [y for y in years for _ in range(len(all_groups))],
    }
    for var in groupby_area_diff:
        combi_dict[var] = all_groups[var].tolist() * len(years)
    all_combinations = pandas.DataFrame(combi_dict)
    # Do a full join to make NA values apparent in order to compute the diff in
    # area or stock later
    df = df.merge(all_combinations, how="outer", on=["year"] + groupby_area_diff)
    df.fillna(0, inplace=True)
    df.sort_values(groupby_area_diff + ["year"], inplace=True)
    # Compute the area diff and check the diff sums to zero
    df["area_diff"] = df.groupby(groupby_area_diff)["area"].transform(
        lambda x: x.diff()
    )
    diff_sum = abs(df.groupby("year")["area_diff"].sum())
    assert all(diff_sum < 100)
    return df


class Sink:
    """Compute the forest carbon sink in living biomass, dead organic matter
    and soil pools

    Here are the methods to load intermediate data frames used in the
    computation of the sink

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['reference'].runners['LU'][-1]

        >>> runner.post_processor.sink.pools
        >>> runner.post_processor.sink.fluxes
        >>> runner.post_processor.sink.nf_soil_stock
        >>> groupby_sink = ["year", "region", "climate", "status"]
        >>> runner.post_processor.sink.emissions_from_deforestation(groupby=groupby_sink)
        >>> runner.post_processor.sink.emissions_from_deforestation(groupby=["year"])
        >>> runner.post_processor.sink.deforestation_deduction
        >>> runner.post_processor.sink.df
        >>> runner.post_processor.sink.df_agg("year")

    Display the number of rows going through land class changes and disturbances

        >>> pools = runner.post_processor.sink.pools.copy()
        >>> selector = pools["time_since_land_class_change"] > 1
        >>> pools["time_since_land_class_change"] = pools["time_since_land_class_change"].astype(str)
        >>> pools.loc[selector, "time_since_land_class_change"] = "greater than one"
        >>> selector = pools["time_since_last_disturbance"] > 1
        >>> pools["time_since_last_disturbance"] = pools["time_since_last_disturbance"].astype(str)
        >>> pools.loc[selector, "time_since_last_disturbance"] = "greater than one"
        >>> selected_cols = ["time_since_land_class_change", "status", "land_class", "time_since_last_disturbance"]
        >>> print(pools.value_counts(selected_cols, sort=False))
        >>> print("\\nOnly stands affected by deforestation disturbances")
        >>> print(pools.query("last_disturbance_type==7").value_counts(selected_cols, sort=False))

    """

    def __init__(self, parent):
        self.parent = parent
        self.runner = parent.runner
        self.combo_name = self.runner.combo.short_name
        self.pools = self.parent.pools
        self.fluxes = self.parent.fluxes
        # Pools and fluxes aggregation parameters
        self.pools_dict = POOLS_DICT.copy()
        self.pools_list = list(
            {item for sublist in self.pools_dict.values() for item in sublist}
        )
        self.fluxes_dict = FLUXES_DICT.copy()
        self.groupby_sink = ["year", "region", "climate", "status"]

    def __repr__(self):
        return '%s object code "%s"' % (self.__class__, self.runner.short_name)

    @cached_property
    def nf_soil_stock(self):
        """Get the slow soil pool content per hectare of non forested stands.

        Used to compute the sink of afforested land in the first year of
        afforestation. Keep only stands that have never been disturbed in the
        simulation (time_since_land_class_change == -1), exclude NF stands that
        are the result of deforestation during the simulation period.
        """
        df = self.pools
        selector = df["status"].str.contains("NF")
        selector &= df["time_since_land_class_change"] == -1
        nf_soil = df.loc[selector].copy()
        nf_soil["nf_slow_soil_per_ha"] = (
            nf_soil["below_ground_slow_soil"] / nf_soil["area"]
        )
        # Group by region and climate and calculate the standard deviation
        groupby_soil = ["region", "climate"]
        nf_soil["std_dev"] = nf_soil.groupby(groupby_soil)[
            "nf_slow_soil_per_ha"
        ].transform("std")
        # Check that nf_slow_soil_per_ha always have the same value across grouping
        # variables
        selector = nf_soil["std_dev"] > 1e-2
        if any(selector):
            msg = "The NF non forested soil pool content per hectare"
            msg += " is not homogeneous for some region and climate groups."
            cols_to_show = ["year", "status", "region", "climate"]
            cols_to_show += [
                "time_since_land_class_change",
                "nf_slow_soil_per_ha",
                "std_dev",
            ]
            msg += f"{nf_soil[cols_to_show]}"
            raise ValueError(msg)
        # Aggregate smaller data frame with columns necessary for the join
        nf_soil_agg = nf_soil.groupby(groupby_soil)["nf_slow_soil_per_ha"].agg("mean")
        nf_soil_agg = nf_soil_agg.reset_index()
        return nf_soil_agg

    def emissions_from_deforestation(
        self,
        groupby: Union[List[str], str],
        fluxes_dict: dict = None,
        current_year_only: bool = False,
    ):
        """Emissions from deforested areas moving from forested to NF

        Deforestation emissions are only reported for the year when event happens.
        Indeed, a small amount of legacy emissions occur, as reflected by
        "decay_domco2_emission"evolution after deforestation for any identifier We
        considered it as nonrelevant, anyway atributable to post-deforestation land
        use. Deforestation emissions can be identified by dist_type = 7, OR,
        "status = "NF" and "time_since_land_class_change > 0" land transfers from
        occur ForAWS/NAWS, or even AR, to NF. join this df to sink df_all.

        Example use:

            >>> from eu_cbm_hat.core.continent import continent
            >>> runner = continent.combos['reference'].runners['LU'][-1]
            >>> runner.post_processor.sink.emissions_from_deforestation(groupby=["year"])
            >>> runner.post_processor.sink.emissions_from_deforestation(groupby=["year", "region"])

            >>> # TODO update this example
            >>> def_em_y = apply_to_all_countries(emissions_from_deforestation, combo_name="reference", groupby="year")

        """
        if fluxes_dict is None:
            fluxes_dict = self.fluxes_dict.copy()
        df = self.fluxes.copy()
        # Keep only deforestation events
        selector = df["time_since_land_class_change"] > 0
        selector &= df["last_disturbance_type"] == 7
        # Used to compute the deforestation deduction for the current year only
        # when computing the sink
        if current_year_only:
            selector &= df["time_since_last_disturbance"] == 1
            df = df.loc[selector]

        for key in fluxes_dict:
            # Aggregate all pool columns to one pool value for this key
            df[key] = df[fluxes_dict[key]].sum(axis=1)

        cols = list(fluxes_dict.keys())
        # Aggregate
        df_agg = df.groupby(groupby)[cols].agg("sum").reset_index()

        return df_agg

    @cached_property
    def deforestation_deduction(self):
        """Pool content and fluxes from the area subject to deforestation Prepare a
        data frame to deduce carbon related to deforestation"""
        # Aggregate by the classifier for which it is possible to compute a
        # difference in pools.
        df = self.pools
        selector = df["last_disturbance_type"] == 7
        selector &= df["time_since_last_disturbance"] == 1
        df7 = df.loc[selector].copy()
        df7["area_deforested_current_year"] = df["area"]
        selected_columns = self.pools_list + ["area_deforested_current_year"]
        df7_agg = df7.groupby(self.groupby_sink)[selected_columns].sum().reset_index()
        def_em = self.emissions_from_deforestation(
            groupby=self.groupby_sink, fluxes_dict=FLUXES_DICT, current_year_only=True
        )
        deforest = df7_agg.merge(def_em, on=self.groupby_sink, how="outer")
        if deforest.status.unique() != "NF":
            msg = "After deforestation the status should be NF only. "
            msg += f"but it is {deforest.status.unique()}"
            raise ValueError(msg)
        status_foraws = "ForAWS"
        if status_foraws not in df["status"].unique():
            msg = f"{status_foraws} not in df['status']: {df['status'].unique()}"
            raise ValueError(msg)
        # Replace status NF by ForAWS
        deforest["status"] = status_foraws
        # Compute the deforestation deduction
        for key, pools in POOLS_DICT.items():
            deforest[key + "_stock"] = deforest[pools].sum(axis=1)
            col_name = deforest.columns[deforest.columns.str.contains("_from_" + key)][
                0
            ]
            deforest[key + "_deforest_deduct"] = (
                deforest[key + "_stock"] + deforest[col_name]
            )
        return deforest

    @cached_property
    def df(self):
        """Compute the stock change and the sink

        Aggregate by the classifier for which it is possible to compute a
        difference in pools. During land use transition implementing afforestation
        and deforestation, some classifier sets may change, while other classifiers
        such as region and climate remain constant. It is only possible to compute
        the stock change along classifiers that remain constant.

        Normalise the sink by the area. For example in case of afforestation the
        stock change should take into account the change of area from t-1 to t.
        Steps to correct for the area change:

            - Group by ["year", "region", "climate", "status",
                        "afforestation_in_current_year"] and sum pools
            - Aggregate all pool columns to one pool value for each key in the
              pools_dict dictionary
            - Compute the stock change per hectare
                S{t}/A{t} - S{t-1}/A{t-1}
            - Deduce NF soil pool when there is afforestation in the first year
            - Compute the CO2 eq. sink per hectare
            - Multiply the sink by the area at time t
            - Remove non forested land
            - Group by final grouping variables given in the groupby argument.

        See usage example in the function sink_one_country.

        Investigate issues with area changes

            >>> from eu_cbm_hat.core.continent import continent
            >>> runner = continent.combos['reference'].runners['LU'][-1]
            >>> df = runner.post_processor.sink.df
            >>> df["area_diff"] = df["area"] - df["area_tm1"]
            >>> cols = df.columns[df.columns.str.contains("area")].to_list()
            >>> cols += ["year", "region", "climate", "status"]
            >>> df[cols].to_csv("/tmp/lu_area.csv")

        """
        groupby_sink = self.groupby_sink.copy()

        # Aggregate by the classifier for which it is possible to compute a
        # difference in pools.
        selected_columns = self.pools_list.copy()
        selected_columns += [
            "area",
            "area_afforested_current_year",
            "area_deforested_current_year",
        ]
        df = self.pools.groupby(groupby_sink)[selected_columns].sum().reset_index()

        # Add the soil stock in NF stands (that have not been deforested in the
        # simulation)
        df = df.merge(self.nf_soil_stock, on=["region", "climate"], how="left")

        # Aggregate on the groupby_sink columns and fill na before computing the diff
        df = generate_all_combinations_and_fill_na(df, groupby=groupby_sink)

        # Join deforestation deductions to the main sink data frame
        deforest = self.deforestation_deduction
        selected_cols = deforest.columns[
            deforest.columns.str.contains("_deduct")
        ].to_list()
        df = df.merge(
            deforest[groupby_sink + selected_cols], on=groupby_sink, how="left"
        )
        df[selected_cols] = df[selected_cols].fillna(0)

        # Remove year from the grouping variables to compute the diff over years
        groupby_sink.remove("year")

        # Arrange by group variables with year last to prepare for diff() and shift()
        df.sort_values(groupby_sink + ["year"], inplace=True)

        # Add area at {t-1} to compare with area at t
        df["area_tm1"] = df.groupby(groupby_sink)["area"].transform(lambda x: x.shift())

        # Check that the total area didn't change between t-1 and t
        # Note: the status can change but the total area should remain constant
        df_check = df.groupby("year")[["area", "area_tm1"]].agg("sum")
        df_check = df_check[df_check.index > df_check.index.min()]
        try:
            np.testing.assert_allclose(df_check["area_tm1"], df_check["area"], atol=10)
        except AssertionError as error:
            msg = "The total area changed between t-1 and t"
            raise AssertionError(msg) from error

        for key in self.pools_dict:
            # Aggregate all pool columns to one pool value for this key
            df[key + "_stock"] = df[self.pools_dict[key]].sum(axis=1)

            # Keep stock at {t-1} for debugging purposes
            df[key + "_stock_tm1"] = df.groupby(groupby_sink)[key + "_stock"].transform(
                lambda x: x.shift()
            )

            # Compute the stock change per hectare
            # TODO: change the computation of the stock change so that
            # It becomes possible to analyse stock_t, stock_{t-1}
            df[key + "_stk_ch"] = df.groupby(groupby_sink)[key + "_stock"].transform(
                lambda x: x.diff()
            )

            # Remove the NF soil pool content for the area afforested in current year
            if "soil" in key:
                nf_slow_soil = (
                    df["nf_slow_soil_per_ha"] * df["area_afforested_current_year"]
                )
                df[key + "_stk_ch"] = df[key + "_stk_ch"] - nf_slow_soil

            # Remove the deforestation stock and emissions
            df[key + "_stk_ch"] += df[key + "_deforest_deduct"]

            # Compute the CO2 eq. Sink
            df[key + "_sink"] = df[key + "_stk_ch"] * -44 / 12

        # Remove non forested land
        selector = df["status"].str.contains("NF")
        df = df.loc[~selector]
        return df

    def df_agg(self, groupby: Union[List[str], str] = None):
        """Aggregated sink data frame"""
        if isinstance(groupby, str):
            groupby = [groupby]
        if groupby is None:
            groupby = self.groupby_sink.copy()
        if not set(groupby).issubset(self.groupby_sink):
            msg = f"Can only group by {self.groupby_sink}. "
            msg += f"{set(groupby) - set(self.groupby_sink)}"
            msg += " not allowed as a groupby value."
            raise ValueError(msg)

        # Keep the area, pool and sink information
        cols = self.df.columns
        selected_cols = [
            "area",
            "area_afforested_current_year",
            "area_deforested_current_year",
        ]
        selected_cols += cols[cols.str.contains("stock")].to_list()
        selected_cols += cols[cols.str.contains("sink$")].to_list()
        selected_cols += cols[cols.str.contains("deforest_deduct")].to_list()
        # Aggregate selected columns by the final grouping variables
        df_agg = self.df.groupby(groupby)[selected_cols].agg("sum").reset_index()
        return df_agg



"""Conversion functions"""

from eu_cbm_hat import CARBON_FRACTION_OF_BIOMASS


def ton_carbon_to_m3_ub(df, input_var):
    """Convert tons of carbon to volume in cubic meter under bark

    The input data frame must contain the bark_frac and wood_density columns.
    """
    return (df[input_var] * (1 - df["bark_frac"])) / (
        CARBON_FRACTION_OF_BIOMASS * df["wood_density"]
    )


def ton_carbon_to_m3_ob(df, input_var):
    """Convert tons of carbon to volume in cubic meter over bark

    The input data frame must contain the wood_density column.
    """
    return df[input_var] / (
        CARBON_FRACTION_OF_BIOMASS * df["wood_density"]
    )


# addedd for outputs on softwood/con and hardwood/broad 
def ton_carbon_to_m3_ub_soft(df, input_var):
    """Convert tons of carbon to volume in cubic meter under bark

    The input data frame must contain the bark_frac and wood_density columns.
    """
    return (df[input_var] * (1 - df["bark_frac"])) / (
        CARBON_FRACTION_OF_BIOMASS * df["wood_density"]
    )

def ton_carbon_to_m3_ub_hard(df, input_var):
    """Convert tons of carbon to volume in cubic meter under bark

    The input data frame must contain the bark_frac and wood_density columns.
    """
    return (df[input_var] * (1 - df["bark_frac"])) / (
        CARBON_FRACTION_OF_BIOMASS * df["wood_density"]
    )

def ton_carbon_to_m3_ob_soft(df, input_var):
    """Convert tons of carbon to volume in cubic meter over bark

    The input data frame must contain the wood_density column.
    """
    return df[input_var] / (
        CARBON_FRACTION_OF_BIOMASS * df["wood_density"]
    )

def ton_carbon_to_m3_ob_hard(df, input_var):
    """Convert tons of carbon to volume in cubic meter over bark

    The input data frame must contain the wood_density column.
    """
    return df[input_var] / (
        CARBON_FRACTION_OF_BIOMASS * df["wood_density"]
    )



"""Diagnostic tables and plots

To analyse the model output.
"""

class Diagnostic:
    """Diagnostic tables and plots

    Usage:

        >>> from eu_cbm_hat.core.continent import continent
        >>> from matplotlib import pyplot as plt
        >>> runner = continent.combos['reference'].runners['LU'][-1]
        >>> runner.post_processor.diagnostic.plot_n_stands(by="status")
        >>> plt.show()

        >>> runner.post_processor.diagnostic.plot_n_stands(by="forest_type")
        >>> plt.show()

    Unique values of disturbance_type and last_disturbance_type

        >>> fluxes = runner.post_processor.fluxes.copy()
        >>> fluxes["time_since_dist"] = fluxes["time_since_last_disturbance"]
        >>> locator = fluxes["time_since_last_disturbance"] > 1
        >>> fluxes.loc[locator, "time_since_dist"] = "more"
        >>> # Count the number of rows for each unique combination of the selected cols
        >>> cols = ["time_since_dist", "disturbance_type", "last_disturbance_type"]
        >>> df = fluxes.value_counts(cols).reset_index().sort_values(cols)

    Therefore

    - the condition disturbance_type == 7

    - is equivalent to the condition time_since_last_disturbance == 1
      & last_disturbance_type == 7

    And the condition

    - disturbance_type==0

    - is almost equivalent to the condition time_since_last_disturbance >= 2
      except for the special case where disturbance_type==0 and
      last_disturbance_type ==2

    """

    def __init__(self, parent):
        self.parent = parent
        self.runner = parent.runner
        self.combo_name = self.runner.combo.short_name
        self.country_name = self.runner.country.country_name
        self.pools = self.parent.pools
        self.fluxes = self.parent.fluxes

    def n_stands(self, groupby: str):
        """Number of stands in the model along the given grouping variables"""
        df = self.parent.pools
        df = df[[groupby, "year"]].value_counts().reset_index()
        return df

    def plot_n_stands(self, by: str):
        """Plot the number of stands in the model along the given classifier
        variable. Only one column name is allowed as the by variable.
        """
        df = self.n_stands(groupby=by)
        df = df.pivot(index="year", values="count", columns=by)
        title = "Number of stands in "
        title += f"{self.country_name} - {self.combo_name} combo"
        return df.plot(title=title)






#!/usr/bin/env python
# coding: utf-8
# %%
# import FAO and CRF databases needed for HWP estimation. This will include all countries. 

# %%
from typing import Union, List
from functools import cached_property
from eu_cbm_hat import eu_cbm_data_pathlib
from eu_cbm_hat.post_processor.hwp_input import HWPInput
import numpy as np
import pandas as pd


# %%
class HWP(object):
    """Harvest Wood Products input and processing

    Usage:

        >>> from eu_cbm_hat.post_processor.hwp import HWP
        >>> hwp = HWP
        >>> hwp.rw_export_correction_factor

    """

    @cached_property
    def input(self):
        """Input data for HWP computations"""
        return HWPInput(self)

    
    @cached_property
    def rw_export_correction_factor(self):
        """ data 1961-2021 is from Forestry_E_Europe.csv
        this function allows the estimation of the factor "f" that represents the feedstock for the HWP of domestic origin, 
        after the correction for the export of roundwood, to be applied to eu_cbm_hat simulated IRW. 
        Even two types of fractions are calculated, fraction with string '_dom' is used further """
        """runner.post_processor.hwp.rw_export_correction_factor()"""

        # you should make this a method of HWP so that 
        # df_fao = self.faostat_bulk_data
        df_fao = self.input.faostat_bulk_data

        # remove rows which do not reffer to "quantity" from original data
        filter = df_fao['Element'].str.contains('Value')
        df_fao = df_fao[~filter].rename(columns = {'Item':'Item_orig', 'Element':'Element_orig'})

        # add lables used in the hwp scripts
        df = df_fao.merge(self.input.hwp_types, on = ['Item Code','Item_orig']).merge(self.input.eu_member_states, on = ['Area'])

        # Filter the columns that start with 'Y' and do not end with a letter
        keep_columns = ['Area Code', 'Area', 'Item Code','Item_orig', 'Item', 'Element Code', 'Element_orig', 'Unit']
        fao_stat = df.loc[:, keep_columns + df.columns[(df.columns.str.startswith('Y')) &
                                                       ~(df.columns.str.endswith(('F', 'N')))].tolist()]

        # Rename columns to remove 'Y' prefix for the year
        new_columns = {col: col[1:] if col.startswith('Y') else col for col in df.columns}
        fao_stat = fao_stat.rename(columns=new_columns)

        # reorganize table on long format
        fao_stat = fao_stat.melt(id_vars=['Area Code', 'Area', 'Item Code', 'Item_orig', 'Item','Element Code', 'Element_orig',
                                            'Unit'], var_name='year', value_name='Value')
        # add new labels on a new column for harmonization 

        shorts_mapping = {
                    'Production': 'prod',
                    'Import Quantity': 'imp',
                    'Export Quantity': 'exp'}
        fao_stat.loc[:,'Element'] = fao_stat.loc[:,'Element_orig'].map(shorts_mapping)

        #rename
        fao_stat=fao_stat.rename(columns = {'Area':'area'})  

        #aggregate on labels
        df_exp = (fao_stat
                    .groupby(['area', 'Element', 'year', 'Item'])
                    .agg(value = ('Value', 'sum'))
                    .reset_index()
                         )
        # create the input type
        df_exp ['type'] = df_exp ['Item'] .astype(str)+"_"+df_exp ['Element'].astype(str)

        # convert long to wide format
        df_exp= df_exp.pivot(index=['area', 'year'], columns=['type'], values=['value'])
        df_exp=df_exp.droplevel(None, axis=1).reset_index()

        # replacing NA to 0, so possible to make aritmetic operations
        df_exp=df_exp.fillna(0)

        # add con and broad aggregates, for information purpose only as split on con and broad is mantained
        # reduce to IRW
        df_exp['irw_prod'] = df_exp['irw_broad_prod'] + df_exp['irw_con_prod']
        df_exp['irw_exp'] = df_exp['irw_broad_exp'] + df_exp['irw_con_exp']
        df_exp['irw_imp'] = df_exp['irw_broad_imp'] + df_exp['irw_con_imp']

        #df_exp.to_csv('C:/CBM/exp.csv')
        # estimate the fractions of domestic in the country's feedstock: IRW, WP, PULP on con and broad
        #df_exp['fIRW_SW_WP_con'] = (df_exp['irw_con_prod']-df_exp['irw_con_exp'] )/(df_exp['irw_con_prod']+
        #                                                                            df_exp['irw_con_imp'] -
        #                                                                            df_exp['irw_con_exp'] )
        #df_exp['fIRW_SW_WP_broad'] = (df_exp['irw_broad_prod']-df_exp['irw_broad_exp'] )/(df_exp['irw_broad_prod']+
        #                                                                                  df_exp['irw_broad_imp'] - 
        #                                                                                  df_exp['irw_broad_exp'] )
        #df_exp['fPULP'] = (df_exp['wood_pulp_prod']-df_exp['wood_pulp_exp'] )/(df_exp['wood_pulp_prod']+
        #                                                                       df_exp['wood_pulp_imp'] -
        #                                                                       df_exp['wood_pulp_exp'] )

        df_exp['fIRW_SW_WP_con_dom'] = (df_exp['irw_con_prod']-df_exp['irw_con_exp'] )/df_exp['irw_con_prod']
        df_exp['fIRW_SW_WP_broad_dom'] = (df_exp['irw_broad_prod']-df_exp['irw_broad_exp'] )/df_exp['irw_con_prod']
        df_exp['fPULP_dom'] = (df_exp['wood_pulp_prod']-df_exp['wood_pulp_exp'] )/df_exp['irw_con_prod']

        # estimate the generic fraction of domestic feedstock
        df_exp['fIRW_SW_WP'] = (df_exp['irw_prod']-df_exp['irw_exp'] )/(df_exp['irw_prod']+
                                                                        df_exp['irw_imp'] -
                                                                        df_exp['irw_exp'] )

        # apply assumptions that f = 0 or f = 1 when f values are <0 or >1
        # when numerator is negative (export > production)
        # a value f =1 means entire amount corrected result as a domestic feedstock
        # a value f = 0 means there is no domestic contribution to the feedstock  
        #df_exp['fIRW_SW_WP_con'] = df_exp['fIRW_SW_WP_con'].mask(df_exp['fIRW_SW_WP_con']<0, 0)
        #df_exp['fIRW_SW_WP_broad'] = df_exp['fIRW_SW_WP_broad'].mask(df_exp['fIRW_SW_WP_broad']<0, 0)
        #df_exp ['fPULP']= df_exp['fPULP'].mask(df_exp['fPULP']<0, 0)

        df_exp['fIRW_SW_WP_con_dom'] = df_exp['fIRW_SW_WP_con_dom'].mask(df_exp['fIRW_SW_WP_con_dom']<0, 0)
        df_exp['fIRW_SW_WP_broad_dom'] = df_exp['fIRW_SW_WP_broad_dom'].mask(df_exp['fIRW_SW_WP_broad_dom']<0, 0)
        df_exp ['fPULP_dom']= df_exp['fPULP_dom'].mask(df_exp['fPULP_dom']<0, 0)

        # when both numerator and denominatore is negative (export > production & export > production + import)
        #df_exp['fIRW_SW_WP_con'] = df_exp['fIRW_SW_WP_con'].mask(df_exp['fIRW_SW_WP_con']>1, 0)
        #df_exp['fIRW_SW_WP_broad'] = df_exp['fIRW_SW_WP_broad'].mask(df_exp['fIRW_SW_WP_broad']>1, 0)
        #df_exp ['fPULP']= df_exp['fPULP'].mask(df_exp['fPULP']>1, 0)
        #df_exp['fIRW_SW_WP_con'] =df_exp['fIRW_SW_WP_con'].fillna(0)
        #df_exp['fIRW_SW_WP_broad'] =df_exp['fIRW_SW_WP_broad'].fillna(0)
        #df_exp['fPULP'] =df_exp['fPULP'].fillna(0)

        df_exp['fIRW_SW_WP_con_dom'] = df_exp['fIRW_SW_WP_con_dom'].mask(df_exp['fIRW_SW_WP_con_dom']>1, 0)
        df_exp['fIRW_SW_WP_broad_dom'] = df_exp['fIRW_SW_WP_broad_dom'].mask(df_exp['fIRW_SW_WP_broad_dom']>1, 0)
        df_exp ['fPULP_dom']= df_exp['fPULP_dom'].mask(df_exp['fPULP_dom']>1, 0)
        df_exp['fIRW_SW_WP_con_dom'] =df_exp['fIRW_SW_WP_con_dom'].fillna(0)
        df_exp['fIRW_SW_WP_broad_dom'] =df_exp['fIRW_SW_WP_broad_dom'].fillna(0)
        df_exp['fPULP_dom'] =df_exp['fPULP_dom'].fillna(0)

        
        # fractions of recycled paper feedstock, exports and exports
        df_exp['fREC_PAPER'] = (df_exp['recycled_paper_prod']-df_exp['recycled_paper_exp'] )/(df_exp['recycled_paper_prod']+
                                                                                    df_exp['recycled_paper_imp'] -
                                                                                    df_exp['recycled_paper_exp'] )

        #replacing NA to 0, so possible to make operations
        df_exp['fREC_PAPER'] =df_exp['fREC_PAPER'].fillna(0)
        df_exp['year'] =df_exp['year'].astype(int)
        return df_exp   


# %%
#initiate the class
hwp = HWP()

# These intermediate data frames are not needed,
# hwp.input.faostat_bulk_data can be called directly instead
faostat_bulk_data = hwp.input.faostat_bulk_data
hwp_types = hwp.input.hwp_types
eu_member_states = hwp.input.eu_member_states
euwrb_stat = hwp.input.euwrb_stat
fao_rw_prod_gapfilled = hwp.input.fao_rw_prod_gapfilled
crf_stat = hwp.input.crf_stat
subst_params = hwp.input.subst_params
subst_ref = hwp.input.subst_ref
silv_to_hwp = hwp.input.silv_to_hwp


# %%
# TODO remove this legacy function that just mirrors the HWP method
def rw_export_correction_factor():
    """Ugly function which should not exist"""
    return hwp.rw_export_correction_factor


# %%
def crf_semifinished_data():
        """ data 1961-2021 from common\hwp_crf_submission_2023.csv
        input timeseries of quantities of semifinshed products reported under the CRF"""
        df_crf=crf_stat.set_index(['area', 'year'])
        selector = '_crf'
        df_crf=df_crf.filter(regex=selector).reset_index()
        
        # remove strings in names
        df_crf.columns=df_crf.columns.str.replace(selector,'')
        df_crf=df_crf.set_index(['area', 'year'])
        
        # remove notation kew from CRF based data
        df_crf=df_crf.replace (["NO", 'NE', 'NA', 'NA,NE'], 0)
        df_crf=df_crf.fillna(0).astype(float)
        df_crf=df_crf.filter(regex='_prod').reset_index()
        df_crf['year'] =df_crf['year'].astype(int)
        return df_crf


# %%
# these are the historical domestic feedstock (corrected for export)
def hist_domestic_semifinished_production():
        """this merges the export with semifinished inputs to generate HWP of domestic origin, 
        in original unit m3 or t for 1961-2021"""
        
        df_exp = rw_export_correction_factor()
        df_crf = crf_semifinished_data()
        df_dp = df_exp.merge(df_crf, on=['year', 'area'])
        # generic factor is used
        df_dp ['sw_prod'] = df_dp ['sw_prod_m3']*df_dp ['fIRW_SW_WP'].astype(float) 
        df_dp['wp_prod'] = df_dp ['wp_prod_m3']*df_dp ['fIRW_SW_WP'].astype(float)
        #_dom is used
        df_dp ['pp_prod'] = df_dp ['pp_prod_t']*df_dp ['fPULP_dom'].astype(float)
        df_dp =df_dp [['area', 'year','sw_prod', 'wp_prod', 'pp_prod']]
        return df_dp


# %%
#def gap_filling_ms_crf():
def gap_filling_eu_totals():

    """add a EU total excluding the countries with incomplete time series, 
       to be used as proxy for gap filling of missing data by ms
       in original unit m3 or t for 1961-2021"""
    """runner.post_processor.hwp.rw_export_correction_factor()"""
    
    df_dp = hist_domestic_semifinished_production()
    df_dp = df_dp[['year', 'area', 'sw_prod', 'wp_prod', 'pp_prod']]
    df_ms = df_dp.rename (columns = {'sw_prod':'sw_prod_ms','wp_prod':'wp_prod_ms','pp_prod':'pp_prod_ms'})
    
    # Group by Area, Item, and Element
    grouped = df_ms.groupby(['year', 'area'])
   
    complete_groups = grouped.filter(lambda x: 
                                  not ((x['sw_prod_ms'] == 0).any() or 
                                       (x['wp_prod_ms'] == 0).any() or
                                       (x['pp_prod_ms'] == 0).any()))
    
    # Group by Area, Item, and Element and sum the amount
    df_eu = complete_groups.groupby(['year'])[['sw_prod_ms','wp_prod_ms','pp_prod_ms']].sum().reset_index()
    df_eu = df_eu.rename(columns={
        'sw_prod_ms': 'sw_prod_eu',
        'wp_prod_ms': 'wp_prod_eu',
        'pp_prod_ms': 'pp_prod_eu'
    })
    
    df_eu['EU'] = 'EU'
    df_crf_eu = df_eu.merge(df_ms, on = ['year'])
    df_crf_eu = df_crf_eu.sort_values (by = ['year','area'] )
    df_crf_eu.replace(0, np.nan, inplace=True)
    #df_crf_eu.to_csv('C:/CBM/interpolated.csv')
    return df_crf_eu


# %%
def gapfill_hwp_ms_backward(df):
    # Copy the original DataFrame to avoid modifying the original data for 1961-2021
    interpolated_ms = df.copy()
    
    # Calculate the ratio of irw_eu for each row to the next row
    interpolated_ms['sw_ratio'] = interpolated_ms['sw_prod_eu'].shift(-1) / interpolated_ms['sw_prod_eu']
    interpolated_ms['wp_ratio'] = interpolated_ms['wp_prod_eu'].shift(-1) / interpolated_ms['wp_prod_eu']
    interpolated_ms['pp_ratio'] = interpolated_ms['pp_prod_eu'].shift(-1) / interpolated_ms['pp_prod_eu']
    
    # Reset the index to ensure consecutive integers
    interpolated_ms.reset_index(drop=True, inplace=True)
    
    # Reverse the DataFrame to fill missing values in reverse order
    interpolated_ms = interpolated_ms.iloc[::-1]
    
    # Fill missing values in new_irw_ms using the ratio
    for index, row in interpolated_ms.iterrows():
        if pd.isnull(row['sw_prod_ms']):
            next_value = interpolated_ms.at[index + 1, 'new_sw_ms']
            if not pd.isnull(next_value):
                interpolated_ms.at[index, 'new_sw_ms'] = int(next_value / row['sw_ratio'])
            else:
                interpolated_ms.at[index, 'new_sw_ms'] = next_value  # Keep NaN if next value is NaN
        else:
            interpolated_ms.at[index, 'new_sw_ms'] = row['sw_prod_ms']
       
    for index, row in interpolated_ms.iterrows():
        if pd.isnull(row['wp_prod_ms']):
            next_value = interpolated_ms.at[index + 1, 'new_wp_ms']
            if not pd.isnull(next_value):
                interpolated_ms.at[index, 'new_wp_ms'] = int(next_value / row['wp_ratio'])
            else:
                interpolated_ms.at[index, 'new_wp_ms'] = next_value  # Keep NaN if next value is NaN
        else:
            interpolated_ms.at[index, 'new_wp_ms'] = row['wp_prod_ms']
            
    for index, row in interpolated_ms.iterrows():
        if pd.isnull(row['pp_prod_ms']):
            next_value = interpolated_ms.at[index + 1, 'new_pp_ms']
            if not pd.isnull(next_value):
                interpolated_ms.at[index, 'new_pp_ms'] = int(next_value / row['pp_ratio'])
            else:
                interpolated_ms.at[index, 'new_pp_ms'] = next_value  # Keep NaN if next value is NaN
        else:
            interpolated_ms.at[index, 'new_pp_ms'] = row['pp_prod_ms']       
    
         
    #for col in ['sw_ratio', 'wp_ratio', 'pp_ratio']:
    #    for index, row in interpolated_ms.iterrows():
    #        if pd.isnull(row[col]):
    #            next_value = interpolated_ms.at[index + 1, f'new_{col[:-6]}_ms']
    #            if not pd.isnull(next_value):
    #                interpolated_ms.at[index, f'new_{col[:-6]}_ms'] = int(next_value / row[col])
    #            else:
    #                interpolated_ms.at[index, f'new_{col[:-6]}_ms'] = next_value  # Keep NaN if next value is NaN
    #        else:
    #            interpolated_ms.at[index, f'new_{col[:-6]}_ms'] = row[col]
    
    # Reverse the DataFrame back to the original order
    interpolated_ms = interpolated_ms.iloc[::-1]
    c_sw = 0.225
    c_pw = 0.294
    c_pp = 0.450
    interpolated_ms['sw_domestic_tc'] = c_sw * interpolated_ms['new_sw_ms']
    interpolated_ms['wp_domestic_tc'] = c_pw * interpolated_ms['new_wp_ms']
    interpolated_ms['pp_domestic_tc'] = c_pp * interpolated_ms['new_pp_ms']
    
    # Drop the temporary 'ratio' column as it's no longer needed
    columns_to_drop = ['sw_ratio', 'wp_ratio', 'pp_ratio']
    interpolated_ms.drop(columns=columns_to_drop, inplace=True)
    
    # Convert 'new_irw_ms' column to integer
    #interpolated_ms#['new_sw_ms'] = interpolated_sw['new_sw_ms'].astype(int)
   
    return interpolated_ms


# %%
def fao_sw_to_irw ():
    """this estimates the average amount of sawnwood produced as average of 2021 and 2022 """
    """runner.post_processor.hwp.rw_export_correction_factor()"""
    #df_faostat = faostat_bulk_data
   
    # remove rows which do not reffer to "quantity" from original data
    filter = faostat_bulk_data['Element'].str.contains('Value')
    df_fao = faostat_bulk_data[~filter].rename(columns = {'Item':'Item_orig', 'Element':'Element_orig'})
   
    
    # add lables used in the hwp scripts
    df = df_fao.merge(hwp_types, on = ['Item Code','Item_orig']).merge(eu_member_states, on = ['Area'])
    
    # Filter the columns that start with 'Y' and do not end with a letter
    keep_columns = ['Area Code', 'Area', 'Item Code','Item_orig', 'Item', 'Element Code', 'Element_orig', 'Unit']
    df = df.loc[:, keep_columns + df.columns[(df.columns.str.startswith('Y')) & ~(df.columns.str.endswith(('F', 'N')))].tolist()]
    
    
    # Rename columns to remove 'Y' prefix for the year
    new_columns = {col: col[1:] if col.startswith('Y') else col for col in df.columns}
    df = df.rename(columns=new_columns)
    
    df_ms = df.query('Item == "sawnwood_broad" | Item == "sawnwood_con" ').copy()
    
    shorts_mapping = {
                'Production': 'prod',
                'Import Quantity': 'imp',
                'Export Quantity': 'exp'}
    df_ms.loc[:, 'Element_ms'] = df_ms.loc[:,'Element_orig'].map(shorts_mapping)
    
    first_year = 2021
    last_year = 2023
    df_ms = df_ms.loc[:, ['Area', 'Item', 'Element_ms'] + [str(year) for year in range(first_year, last_year)]]
    
    df_ms = pd.melt(df_ms, id_vars=['Area', 'Item', 'Element_ms'], var_name='Year', value_name='Value')
    
    df_ms = df_ms.groupby(['Area', 'Item', 'Element_ms', 'Year']).agg(
                            irw_ms=('Value', 'sum'),
    ).reset_index()
    
    # keep only 2021 and 2022
    df_ms=df_ms.query ('Element_ms == "prod" ')
    df_ms = df_ms.query('Year == "2021" | Year == "2022" ')
    
    average_sw_ms = df_ms.groupby(['Area', 'Item'])['irw_ms'].mean().reset_index()
    
    average_sw_ms = average_sw_ms.pivot(index= 'Area', columns='Item', values='irw_ms').reset_index()
    
    # add the share of sawnwood expected from the final cut
    average_sw_ms['final_cut_share_broad'] = 0.9
    average_sw_ms['final_cut_share_con'] = 0.95
        
    return average_sw_ms


# %%
def fao_wp_to_irw ():
    """this estimates the average amount of sawnwood produced as average of 2021 and 2022 """
    """runner.post_processor.hwp.rw_export_correction_factor()"""
    #df_faostat = faostat_bulk_data
   
    # remove rows which do not reffer to "quantity" from original data
    filter = faostat_bulk_data['Element'].str.contains('Value')
    df_fao = faostat_bulk_data[~filter].rename(columns = {'Item':'Item_orig', 'Element':'Element_orig'})
   
    
    # add lables used in the hwp scripts
    df = df_fao.merge(hwp_types, on = ['Item Code','Item_orig']).merge(eu_member_states, on = ['Area'])
    
    # Filter the columns that start with 'Y' and do not end with a letter
    keep_columns = ['Area Code', 'Area', 'Item Code','Item_orig', 'Item', 'Element Code', 'Element_orig', 'Unit']
    df = df.loc[:, keep_columns + df.columns[(df.columns.str.startswith('Y')) & ~(df.columns.str.endswith(('F', 'N')))].tolist()]
    
    
    # Rename columns to remove 'Y' prefix for the year
    new_columns = {col: col[1:] if col.startswith('Y') else col for col in df.columns}
    df = df.rename(columns=new_columns)
    
    df_ms = df.query(' Item == "wood_panels" ').copy()
    
    shorts_mapping = {
                'Production': 'prod',
                'Import Quantity': 'imp',
                'Export Quantity': 'exp'}
    df_ms.loc[:,'Element_ms'] = df_ms.loc[:,'Element_orig'].map(shorts_mapping)
    
    first_year = 2021
    last_year = 2023
    df_ms = df_ms.loc[:, ['Area', 'Item', 'Element_ms'] + [str(year) for year in range(first_year, last_year)]]
    
    df_ms = pd.melt(df_ms, id_vars=['Area', 'Item', 'Element_ms'], var_name='Year', value_name='Value')
    
    df_ms = df_ms.groupby(['Area', 'Item', 'Element_ms', 'Year']).agg(
                            irw_ms=('Value', 'sum'),
    ).reset_index()
    
    # keep only 2021 and 2022
    df_ms=df_ms.query ('Element_ms == "prod" ')
    df_ms = df_ms.query('Year == "2021" | Year == "2022" ')
    
    average_wp_ms = df_ms.groupby(['Area', 'Item'])['irw_ms'].mean().reset_index()
    
    average_wp_ms = average_wp_ms.pivot(index= 'Area', columns='Item', values='irw_ms').reset_index()
    
        
    return average_wp_ms


# %%
def fao_pulp_to_irw ():
    """this estimates the average amount of sawnwood produced as average of 2021 and 2022 """
    """runner.post_processor.hwp.rw_export_correction_factor()"""
    #df_faostat = faostat_bulk_data
   
    # remove rows which do not reffer to "quantity" from original data
    filter = faostat_bulk_data['Element'].str.contains('Value')
    df_fao = faostat_bulk_data[~filter].rename(columns = {'Item':'Item_orig', 'Element':'Element_orig'})
   
    
    # add lables used in the hwp scripts
    df = df_fao.merge(hwp_types, on = ['Item Code','Item_orig']).merge(eu_member_states, on = ['Area'])
    
    # Filter the columns that start with 'Y' and do not end with a letter
    keep_columns = ['Area Code', 'Area', 'Item Code','Item_orig', 'Item', 'Element Code', 'Element_orig', 'Unit']
    df = df.loc[:, keep_columns + df.columns[(df.columns.str.startswith('Y')) & ~(df.columns.str.endswith(('F', 'N')))].tolist()]
    
    
    # Rename columns to remove 'Y' prefix for the year
    new_columns = {col: col[1:] if col.startswith('Y') else col for col in df.columns}
    df = df.rename(columns=new_columns)
    
    df_ms = df.query('Item == "wood_pulp" ').copy()
    
    shorts_mapping = {
                'Production': 'prod',
                'Import Quantity': 'imp',
                'Export Quantity': 'exp'}
    df_ms.loc[:,'Element_ms'] = df_ms.loc[:,'Element_orig'].map(shorts_mapping)
    
    first_year = 2021
    last_year = 2023
    df_ms = df_ms.loc[:, ['Area', 'Item', 'Element_ms'] + [str(year) for year in range(first_year, last_year)]]
       
    df_ms = pd.melt(df_ms, id_vars=['Area', 'Item', 'Element_ms'], var_name='Year', value_name='Value')
    
    df_ms = df_ms.groupby(['Area', 'Item', 'Element_ms', 'Year']).agg(
                            irw_ms=('Value', 'sum'),
    ).reset_index()
    
    # keep only 2021 and 2022
    df_ms=df_ms.query ('Element_ms == "prod" ')
    df_ms = df_ms.query('Year == "2021" | Year == "2022" ')
    
    average_pulp_ms = df_ms.groupby(['Area', 'Item'])['irw_ms'].mean().reset_index()
    
    average_pulp_ms = average_pulp_ms.pivot(index= 'Area', columns='Item', values='irw_ms').reset_index()
        
    return average_pulp_ms


# %%
def substitution_factors ():
        "this merges the export with seminished inputs to generate HWP of domestic origin"
        df_subst_factors = subst_params.rename(columns={'Area': 'area'}) 
        return df_subst_factors


# %%
def gap_filling_irw_faostat():
    """this function allows the gapfilling of production, import and export of roundwood based on FAOSTAT data, 
    i.e., applied to the a quantity result in the domestic amount"""
    """runner.post_processor.hwp.rw_export_correction_factor()"""
    #df_faostat = faostat_bulk_data
   
    # remove rows which do not reffer to "quantity" from original data
    filter = faostat_bulk_data['Element'].str.contains('Value')
    df_fao = faostat_bulk_data[~filter].rename(columns = {'Item':'Item_orig', 'Element':'Element_orig'})
   
    
    # add lables used in the hwp scripts
    df = df_fao.merge(hwp_types, on = ['Item Code','Item_orig']).merge(eu_member_states, on = ['Area'])
    
    # Filter the columns that start with 'Y' and do not end with a letter
    keep_columns = ['Area Code', 'Area', 'Item Code','Item_orig', 'Item', 'Element Code', 'Element_orig', 'Unit']
    df = df.loc[:, keep_columns + df.columns[(df.columns.str.startswith('Y')) & ~(df.columns.str.endswith(('F', 'N')))].tolist()]
    
    
    # Rename columns to remove 'Y' prefix for the year
    new_columns = {col: col[1:] if col.startswith('Y') else col for col in df.columns}
    df = df.rename(columns=new_columns)
   
    df_ms = df.query('Item == "irw_broad" | Item == "irw_con" ').copy()
    
    shorts_mapping = {
                'Production': 'prod',
                'Import Quantity': 'imp',
                'Export Quantity': 'exp'}
    df_ms.loc[:,'Element_ms'] = df_ms.loc[:,'Element_orig'].map(shorts_mapping)
    
    df_ms = df_ms[[ 'Area', 'Item', 'Element_ms','1961', '1962', '1963', '1964', '1965', '1966',
       '1967', '1968', '1969', '1970', '1971', '1972', '1973', '1974', '1975',
       '1976', '1977', '1978', '1979', '1980', '1981', '1982', '1983', '1984',
       '1985', '1986', '1987', '1988', '1989', '1990', '1991', '1992', '1993',
       '1994', '1995', '1996', '1997', '1998', '1999', '2000', '2001', '2002',
       '2003', '2004', '2005', '2006', '2007', '2008', '2009', '2010', '2011',
       '2012', '2013', '2014', '2015', '2016', '2017', '2018', '2019', '2020',
       '2021', '2022']]
    
    df_ms = pd.melt(df_ms, id_vars=['Area', 'Item', 'Element_ms'], var_name='Year', value_name='Value')
    
    df_ms = df_ms.groupby(['Area', 'Item', 'Element_ms', 'Year']).agg(
                            irw_ms=('Value', 'sum'),
    ).reset_index()
    
    #df_ms.to_csv('C:/CBM/df_ms.csv')
    
    # Group by Area, Item, and Element
    grouped = df_ms.groupby(['Area', 'Item', 'Element_ms'])

    # Filter out groups where any value is zero within the time series
    complete_groups = grouped.filter(lambda x: not (x['irw_ms'] == 0).any())

    # Group by Area, Item, and Element and sum the amount
    df_eu = complete_groups.groupby(['Item', 'Element_ms', 'Year'])['irw_ms'].sum().reset_index()
    df_eu =df_eu.rename(columns = {'irw_ms':'irw_eu', 'Element_ms':'Element_eu'})
    
    df_faostat = df_ms.merge(df_eu, on = ['Item','Year'])
    #df_faostat = df_faostat.sort_values (by = ['Area','Item', 'Element', 'Year'] )
    df_faostat.replace(0, np.nan, inplace=True)
        
    #df_faostat.to_csv('C:/CBM/df_faostat.csv')
    return df_faostat


# %%
def gapfill_irw_ms_backward(df):
    # Copy the original DataFrame to avoid modifying the original data
    interpolated_df = df.copy()
    interpolated_df=interpolated_df.sort_values(by = ['Year'], ascending=True).sort_values(by = ['Item','Element_ms'])
    
    # Calculate the ratio of irw_eu for each row to the next row
    interpolated_df['ratio'] = interpolated_df['irw_eu'].shift(-1) / interpolated_df['irw_eu']
    
    # Reset the index to ensure consecutive integers
    interpolated_df.reset_index(drop=True, inplace=True)
    
    # Reverse the DataFrame to fill missing values in reverse order
    interpolated_df = interpolated_df.iloc[::-1]
    
    # Fill missing values in new_irw_ms using the ratio
    for index, row in interpolated_df.iterrows():
        if pd.isnull(row['irw_ms']):
            next_value = interpolated_df.at[index + 1, 'new_irw_ms']
            if not pd.isnull(next_value):
                interpolated_df.at[index, 'new_irw_ms'] = int(next_value / row['ratio'])
            else:
                interpolated_df.at[index, 'new_irw_ms'] = next_value  # Keep NaN if next value is NaN
        else:
            interpolated_df.at[index, 'new_irw_ms'] = row['irw_ms']
    
    # Reverse the DataFrame back to the original order
    interpolated_df = interpolated_df.iloc[::-1]
    
    # Drop the temporary 'ratio' column as it's no longer needed
    interpolated_df.drop(columns=['ratio'], inplace=True)
    
    # Convert 'new_irw_ms' column to integer
    interpolated_df['new_irw_ms'] = interpolated_df['new_irw_ms'].astype(int)
    #interpolated_df.to_csv('C:/CBM/interpolated.csv')
    
    return interpolated_df


# %%
def eu_wrb():
    
    sankey_rw_prod_in = euwrb_stat
    
    #retain only rows relevant for production of hwp
    sankey_rw_prod_in_hwp=sankey_rw_prod_in[sankey_rw_prod_in['label']!= 'trade']
    
    #load quantities from orih=ginal database
    sankey_rw_prod_in_hwp= sankey_rw_prod_in_hwp[['scenario','country','year', 'label', 'data']].dropna()
    
    sankey_rw_prod_in_hwp =sankey_rw_prod_in_hwp.pivot(index=['scenario','country','year'], columns='label', values='data').reset_index()
   
    sankey_rw_prod_in_hwp=sankey_rw_prod_in_hwp.reset_index()
    
    #sankey_rw_prod_in_hwp.to_csv('C:/CBM/hwp.csv')   
    
    # the sum of all semifinshed products converted to C
    c_sw = 0.225
    c_pw = 0.294
    c_pp = 0.450
    wd_con_broad = 0.5
    c_fraction = 0.5
    sankey_rw_prod_in_hwp['c_hwp_fao'] = 1*(sankey_rw_prod_in_hwp['pan_ind2fibboa']*c_pw+
                                            sankey_rw_prod_in_hwp['pan_ind2partboa']*c_pw+
                                            sankey_rw_prod_in_hwp['pan_ind2plyven']*c_pw+
                                            sankey_rw_prod_in_hwp['rw4mat2pu_ind']*c_pp+
                                            sankey_rw_prod_in_hwp['saw_ind2sawnw']*c_sw)
    
    #keep only the total C in all semifinished products
    sankey_c_semifinshed_faostat = sankey_rw_prod_in_hwp[['country','year','c_hwp_fao' ]]
    
    #extract the share of sawnwood in total solid production
    sankey_rw_prod_in_hwp['fSW']  =  sankey_rw_prod_in_hwp['saw_ind2sawnw']/sankey_rw_prod_in_hwp['rw_tot2rw4mat']
    
    # production of paper in total solid production, original excel formula: PP = if(rw4mat2pu_ind = 0,0, pu4pa2pap_ind)
    sankey_rw_prod_in_hwp['fPP']=sankey_rw_prod_in_hwp['rw4mat2pu_ind']/sankey_rw_prod_in_hwp['rw_tot2rw4mat']
    
    # deduct the inflow of recycled wood
    # 1st step, define the share of fibboa and partboa in their total
    sankey_rw_prod_in_hwp['fibboa_share'] = sankey_rw_prod_in_hwp['pan_ind2fibboa']/(sankey_rw_prod_in_hwp['pan_ind2fibboa']+sankey_rw_prod_in_hwp['pan_ind2partboa'])
    sankey_rw_prod_in_hwp['partboa_share'] = sankey_rw_prod_in_hwp['pan_ind2partboa']/(sankey_rw_prod_in_hwp['pan_ind2fibboa']+sankey_rw_prod_in_hwp['pan_ind2partboa'])
    
    # 2nd step, further split the total rec wood panels on the two destinations, fibboa and partboa
    sankey_rw_prod_in_hwp['Qfibboa']= sankey_rw_prod_in_hwp['pcw2pan_ind'] * sankey_rw_prod_in_hwp['fibboa_share']
    sankey_rw_prod_in_hwp['Qpartboa']= sankey_rw_prod_in_hwp['pcw2pan_ind'] * sankey_rw_prod_in_hwp['partboa_share']
    
    # 3rd step, estimate production of panels from domestic roundwood, excluding PWC feedstock
    sankey_rw_prod_in_hwp['wp_sum']=sankey_rw_prod_in_hwp['pan_ind2fibboa'] + (sankey_rw_prod_in_hwp['pan_ind2partboa']-sankey_rw_prod_in_hwp['Qpartboa'])+ (sankey_rw_prod_in_hwp['pan_ind2plyven']-sankey_rw_prod_in_hwp['Qfibboa'])
    sankey_rw_prod_in_hwp['wp_sum']
    
    #finally estimate the share of WP in total solid production
    sankey_rw_prod_in_hwp['fWP'] = sankey_rw_prod_in_hwp['wp_sum']/sankey_rw_prod_in_hwp['rw_tot2rw4mat']
    
    
    #sankey_rw_prod_in_hwp
    # # keep the use of post_consumer wood in panels, as it is used in substitution later
    #shares of panels types within WP
    sankey_rw_prod_in_hwp['fWP_fibboa'] = sankey_rw_prod_in_hwp['pan_ind2fibboa']/sankey_rw_prod_in_hwp['wp_sum']
    sankey_rw_prod_in_hwp['fWP_partboa'] = sankey_rw_prod_in_hwp['pan_ind2partboa']/sankey_rw_prod_in_hwp['wp_sum']
    sankey_rw_prod_in_hwp['fWP_pv'] = sankey_rw_prod_in_hwp['pan_ind2plyven']/sankey_rw_prod_in_hwp['wp_sum']
    
    # load recyled wood and paper
    sankey_rw_prod_in_hwp['rec_wood_swe_m3']=sankey_rw_prod_in_hwp['pcw2pan_ind']
    sankey_rw_prod_in_hwp['rec_paper_swe_m3']=sankey_rw_prod_in_hwp['recpap2pap_ind']
    
    # load the export of roundwood
    #retain only rows relevant for production of hwp
    
    
    
    sankey_rw_prod_in_exp=sankey_rw_prod_in[(sankey_rw_prod_in['label'] == 'rw_tot2rw4mat') | (sankey_rw_prod_in['label'] == 'pu4pa2pap_ind')]
    sankey_rw_prod_in_exp=sankey_rw_prod_in_exp[['scenario','country','year', 'data', 'unit', 'label']]
    sankey_rw_prod_in_exp=sankey_rw_prod_in_exp.pivot(index= ['scenario','country','year'] , columns='label' , values='data')
    sankey_rw_prod_in_exp ['rw_export'] =  sankey_rw_prod_in_exp['rw_tot2rw4mat']+sankey_rw_prod_in_exp['pu4pa2pap_ind']

    #sankey_rw_prod_in_exp
    #assess_shares = sankey_rw_prod_in_hwp[['fSW', 'fWP','fPP']]
    
    # complete the db with needed information 
    #fSW = assess_shares.loc['cumulated_values', 'fSW']
    #fWP = assess_shares.loc['cumulated_values', 'fWP']
    #fPP = assess_shares.loc['cumulated_values', 'fPP']
    #fFW = assess_shares.loc['cumulated_values', 'fFW']
    #rec_wood_swe_m3 = sankey_rw_prod_in_hwp.rec_wood_swe_m3.mean()
    #rec_paper_swe_m3 = sankey_rw_prod_in_hwp.rec_paper_swe_m3.mean()
    
    # add further data needed for recycling later
    fWP_fibboa = sankey_rw_prod_in_hwp['fWP_fibboa'].mean()
    fWP_partboa = sankey_rw_prod_in_hwp['fWP_partboa'].mean()
    fWP_pv = sankey_rw_prod_in_hwp['fWP_pv'].mean()
    
    #reorganize 
    sankey_rw_prod = sankey_rw_prod_in_hwp[['scenario','country','year','fSW','fPP','fWP','fWP_fibboa','fWP_partboa','fWP_pv','rec_wood_swe_m3','rec_paper_swe_m3']].copy()
   
    #add absolute amounts of export of roundwood
    rw_export_thou_swe_m3= sankey_rw_prod_in_exp.rw_export.mean()
    sankey_rw_prod.loc[:, 'rw_export_thou_swe_m3'] = rw_export_thou_swe_m3
    
    # convert Sankey volume data to carbon, by taking into account the share of con to broad, and wd
    sankey_rw_prod['rw_export_tc']= 1000*wd_con_broad*c_fraction*sankey_rw_prod['rw_export_thou_swe_m3']

    # estimate the correction
    #retain only rows relevant with annual data on irw_amount allocated to products
    sankey_irw_solid=sankey_rw_prod_in[sankey_rw_prod_in['label'] == 'rw_tot2rw4mat']
    #sankey_irw_solid
    
    return sankey_rw_prod


# %%
def silv_grouping_to_hwp():
    Silv_grouping_to_hwp = silv_to_hwp
    Silv_grouping_to_hwp=Silv_grouping_to_hwp.rename(columns = {'dist_type_name':'disturbance_type'})
    return Silv_grouping_to_hwp



"""Plotting functions

Some of these functions require the model run for all EU countries for many
scenarios to have been aggregated inside post_processor/agg_combos.py.

Usage:

    >>> from eu_cbm_hat.plot.lulucf_plots import plot_sink_by_country
    >>> from eu_cbm_hat.plot.scenarios import plot_hexprov
    >>> from eu_cbm_hat.plot.scenarios import plot_harvest_demand
    >>> from eu_cbm_hat.plot.scenarios import plot_nai, plot_nai_eu

The arguments to these plotting functions are likely to change over time. These
plotting functions where originally created in a notebook at
eu_cbm_explore/scenarios/ssp2_fair_degrowth/ssp2_fair_owc.ipynb

Define a palette to use these plots


"""
import re
import pandas as pd
import seaborn
import seaborn.objects as so
import seaborn as sns
import matplotlib.pyplot as plt


# Rename pathways for the paper
def rename_combo_to_pathway(combo_name):
    """Rename a combo to a pathway
    For example:
        >>> rename_combo_to_pathway("pikssp2_owc_min")
    """
    out = re.sub("pik|_fel1", "", combo_name)
    #out = re.sub("min", "l", out)
    #out = re.sub("max", "h", out)
    return out

def plot_sink_by_country_old (df, y, col_wrap=None, palette=None):
    """Facet plot of CO2 forest sink by country"""
    if col_wrap is None:
        col_wrap = round(len(df["country"].unique()) / 9) + 1
    df = df.copy()
    df[y + "mt"] = df[y] / 1e6
    g = seaborn.relplot(
        data=df,
        x="year",
        y=y + "mt",
        col="country",
        hue="pathway",
        kind="line",
        col_wrap=col_wrap,
        palette=palette,
        facet_kws={"sharey": False, "sharex": False},
    )
    g.set_titles(row_template="{row_name}", col_template="{col_name}")  # , size=30)
    g.fig.set_size_inches(20, 15)
    g.fig.subplots_adjust(hspace=0.3, top=0.95)
    g.set_ylabels(f"{y} MtCO2 eq")
    g.fig.suptitle(f"{y}")
    return g

def plot_sink_by_country(df, y, col_wrap=None, palette=None):
    """Facet plot of CO2 forest sink by country"""
    if col_wrap is None:
        col_wrap = round(len(df["country"].unique()) / 9) + 1
    df = df.copy()
    df[y + "mt"] = df[y] / 1e6

    # Set global font size and line width without grids and with a clean background
    sns.set_theme(style='white', font_scale=1.5, rc={'lines.linewidth': 2.5})

    g = sns.relplot(
        data=df,
        x="year",
        y=y + "mt",
        col="country",
        hue="pathway",
        kind="line",
        col_wrap=col_wrap,
        palette=palette,
        facet_kws={"sharey": False, "sharex": True},
        linewidth=2.5  # Set the line thickness
    )
    
    # Remove the y-axis labels for each subplot
    for ax in g.axes.flat:
        ax.set_ylabel('')
    
    # Set the size of the figure
    g.fig.set_size_inches(20, 15)
    
    # Set a single Y-axis label for the entire figure
    g.fig.text(0.01, 0.5, f"{y} (MtCO2 eq)", va='center', rotation='vertical', fontsize=16)

    # Remove grid lines
    for ax in g.axes.flat:
        ax.grid(False)

    g.fig.suptitle(f"{y}", fontsize=20)

    # Adjust the subplot parameters to give more space for the legend
    g.fig.subplots_adjust(bottom=0.1, top=0.95, hspace=0.3)  # Adjust the top and bottom

    # Draw the legend
    g._legend.remove()
    g.fig.legend(loc='lower center', bbox_to_anchor=(0.5, -0.05), ncol=3)

    return g


def plot_hexprov(df, y, col_wrap=None, palette=None):
    """Facet plot of harvest demand per country for a given product"""
    if col_wrap is None:
        col_wrap = round(len(df["country"].unique()) / 9) + 1
    g = seaborn.relplot(
        data=df,
        x="year",
        y=y,
        col="country",
        hue="pathway",
        style="element",
        kind="line",
        col_wrap=col_wrap,
        palette=palette,
        facet_kws={"sharey": False, "sharex": False},
    )
    g.set(xticks=[2010, 2030, 2050, 2070])
    g.fig.subplots_adjust(top=0.95)
    g.set_titles(row_template="{row_name}", col_template="{col_name}")
    g.fig.set_size_inches(20, 15)
    g.fig.subplots_adjust(hspace=0.3)
    return g

import matplotlib.pyplot as plt

def plot_hexprov_regions(df, y, col_wrap=None, palette=None):
    """Facet plot of harvest demand per country for a given product"""
    if col_wrap is None:
        col_wrap = round(len(df["region_name"].unique()) / 9) + 1

    # Replace NA values with a specific color (e.g., white)
    #df_filtered= df.fillna(value={'year': 0, y: 0})  # Replace NA with 0 or any other value
    df_filtered= df[df["year"]>= 2020]
    
    g = sns.relplot(
        data=df_filtered,
        x="year",
        y=y,
        col="region_name",
        hue="pathway",
        style="element",
        kind="line",
        col_wrap=col_wrap,
        palette=palette,
        facet_kws={"sharey": False, "sharex": False}
    )
    g.set(xticks=[2020, 2030, 2050, 2070])
    g.fig.subplots_adjust(top=0.95)
    g.set_titles(row_template="{row_name}", col_template="{col_name}")
    g.fig.set_size_inches(20, 15)
    g.fig.subplots_adjust(hspace=0.3)

    # Customize the plot to represent NA values as white color
    for ax in g.axes.flat:
        ax.plot([2020, 2070], [0, 0], color="white", linestyle="--")  # Replace NA values with white color

    return g

def plot_hexprov_irw_fw(df, col_wrap=None, palette=None):
    """Facet plot of harvest demand per country for a given product"""
    if col_wrap is None:
        col_wrap = round(len(df["country"].unique()) / 9) + 1

    # Since 'harvest_demand_irw' and 'harvest_demand_fw' are not separate columns,
    # we will assume that the 'harvest' column contains these values.
    df_filtered = df[df['element'].isin(['harvest_demand_irw', 'harvest_demand_fw'])]

    g_ms = sns.relplot(
    data=df_filtered,
    x="year",
    y="harvest",
    col="country",
    hue="pathway",
    style="element",
    kind="line",
    col_wrap=3,
    facet_kws={"sharey": False, "sharex": False},
    )
    g_ms.set(xticks=[2010, 2030, 2050, 2070])
    g_ms.fig.subplots_adjust(top=0.95)
    g_ms.set_titles(row_template="{row_name}", col_template="{col_name}")
    g_ms.fig.set_size_inches(20, 15)
    g_ms.fig.subplots_adjust(hspace=0.3)
    return g_ms


def plot_harvest_demand(df, palette=None):
    """Facet plot of harvest demand per country for a given product"""
    col_wrap = round(len(df["country"].unique()) / 9) + 1
    g = seaborn.relplot(
        data=df,
        x="year",
        y="demand",
        col="country",
        hue="combo_name",
        style="faostat_name",
        kind="line",
        col_wrap=col_wrap,
        palette=palette,
        facet_kws={"sharey": False, "sharex": False},
    )
    g.fig.subplots_adjust(top=0.95)
    g.fig.suptitle(f"Industrial roundwood harvest demand from the economic model")
    g.set_titles(row_template="{row_name}", col_template="{col_name}")
    g.fig.set_size_inches(20, 15)
    g.fig.subplots_adjust(hspace=0.3)
    return g


def plot_nai(df, y, ylabel, forest_type=None, palette=None):
    """Plot Net Annual Increment"""
    g = seaborn.relplot(
        data=df,
        x="year",
        y=y,
        col="country",
        hue="pathway",
        kind="line",
        col_wrap=5,
        palette=palette,
        facet_kws={"sharey": False, "sharex": False},
    )
    status = df["status"].unique()[0]
    title = f"Net Annual Increment {status}"
    if forest_type is not None:
        title += f" forest type: {forest_type}"
    g.fig.suptitle(title)
    g.set_titles(row_template="", col_template="{col_name}")
    g.fig.supylabel(ylabel)
    g.fig.set_size_inches(15, 15)
    g.fig.subplots_adjust(hspace=0.3)
    g.fig.subplots_adjust(top=0.90, left=0.08, right=0.88)
    file_name = f"nai_{y}_{status}_by_country.png"
    return g


def plot_nai_eu(df, y, palette=None):
    """Plot Net Annual Increment in the EU"""
    g = seaborn.relplot(
        data=df,
        x="year",
        y=y,
        col="status",
        hue="pathway",
        kind="line",
        col_wrap=1,
        palette=palette,
        facet_kws={"sharey": False, "sharex": False},
    )
    g.set_titles(row_template="{row_name}", col_template="{col_name}")
    g.fig.supylabel("NAI in million m3")
    g.fig.set_size_inches(12, 10)
    g.fig.subplots_adjust(hspace=0.3)
    g.fig.subplots_adjust(top=0.90, right=0.88)
    return g


def plot_sink_composition(df, selected_years, index):
    """Bar plot of the sink composition for all years
    Facet on pathway, with years on the x axis.
        >>> plot_sink_composition(sink_by_country_groups,
        ...                       [2030, 2050, 2070],
        ...                       index = ["pathway", "year", "country_group"])
        >>> plot_sink_composition(sink_eu,
        ...                       [2030, 2050, 2070],
        ...                       index=["pathway", "year"])
    """
    # Compute dom sink and select years
    df["dom_sink"] = df[["litter_sink", "dead_wood_sink"]].sum(axis=1)
    selected_columns = ["living_biomass_sink", "dom_sink", "soil_sink", "hwp_sink_bau"]
    selector = df["year"].isin(selected_years)
    df = df.loc[selector, index + selected_columns].copy()
    df[selected_columns] = df[selected_columns] / 1e6
    # Reshape to long format
    df_long = df.melt(id_vars=index, var_name="sink", value_name="value")
    # Plot
    p = so.Plot(df_long, x="year", y="value", color="sink")
    p = p.add(so.Bar(), so.Stack())
    if "country_group" in index:
        p = p.facet("pathway", "country_group").share(x=False)
        p = p.layout(size=(14, 9), engine="tight")
    else:
        p = p.facet("pathway").share(x=False)
        p = p.layout(size=(10, 8), engine="tight")
    palette = {
        "living_biomass_sink": "forestgreen",
        "dom_sink": "gold",
        "soil_sink": "black",
        "hwp_sink_bau": "chocolate",
    }
    p = p.scale(x=so.Continuous().tick(at=selected_years), color=palette)
    p = p.label(x="", y="Million t CO2 eq", color="")
    return p


def plot_sink_by_regions(df, y, col_wrap=None, palette=None):
    """Facet plot of harvest demand per country for a given product"""
    if col_wrap is None:
        col_wrap = round(len(df["region_name"].unique()) / 9) + 1

    # Replace NA values with a specific color (e.g., white)
    df_filled = df.fillna(value={'year': 0, y: 0})  # Replace NA with 0 or any other value

    g = sns.relplot(
        data=df_filled,
        x="year",
        y=y,
        col="region_name",
        hue="pathway",
        kind="line",
        col_wrap=col_wrap,
        palette=palette,
        facet_kws={"sharey": False, "sharex": False}
    )
    g.set(xticks=[2010, 2030, 2050, 2070])
    g.set_titles(row_template="{row_name}", col_template="{col_name}")
    g.fig.set_size_inches(20, 15)
    g.fig.subplots_adjust(hspace=0.3)

    # Remove grid lines
    for ax in g.axes.flat:
        ax.grid(False)

    # Adjust the subplot parameters to reduce space on the right and below
    g.fig.subplots_adjust(left=0.08, bottom=0.08, top=0.92, right=0.92)

    # Draw the legend and place it below the plots
    g._legend.remove()
    g.fig.legend(loc='lower center', bbox_to_anchor=(0.5, -0.05), ncol=3)

    g.fig.suptitle(f"{y}", fontsize=20)

    return g



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #
import shutil
import tempfile

# Third party modules #
import pandas

# First party modules #
from plumbing.databases.sqlite_database import SQLiteDatabase

# Internal modules #
from eu_cbm_hat.pump.long_or_wide import events_wide_to_long


###############################################################################
class PreProcessor(object):
    """
    This class will update the input data of a runner based on a set of rules.

    - Load the modified disturbance matrix from
      `eu_cbm/eu_cbm_data/countries/AT/silv/` for the scenario defined in
      pikssp2_owc_max and change the AIDB for the given disturbance matrix
      rows:

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner  = continent.combos['pikssp2_owc_max'].runners['AT'][-1]
        >>> runner.silv.check()
        >>> runner.silv.dist_matrix_value.df
        >>> runner.pre_processor.copy_and_change_aidb()

    """

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        self.runner = parent
        self.country = parent.country
        # Shortcuts #
        self.input = self.runner.input_data

    def __repr__(self):
        return '%s object code "%s"' % (self.__class__, self.runner.short_name)

    # --------------------------- Special Methods -----------------------------#
    def __call__(self):
        # Message #
        self.parent.log.info("Pre-processing input data.")
        # Check empty lines in all CSV inputs #
        for csv_path in self.all_csv:
            self.raise_empty_lines(csv_path)
        # Reshape the events file #
        self.reshape_events()
        # Check there are no negative timesteps #
        self.raise_bad_timestep()
        # Copy the AIDB
        # In case the scenario combination changes the disturbance matrix
        self.copy_and_change_aidb()

    # ----------------------------- Properties --------------------------------#
    @property
    def all_csv(self):
        """Get all CSV inputs in a list."""
        return [
            item.path_obj
            for item in self.input.paths._paths
            if item.path_obj.extension == "csv"
        ]

    # ------------------------------- Methods ---------------------------------#
    @staticmethod
    def raise_empty_lines(csv_path):
        """
        Loads one CSV files and raise an exception if there are any empty
        lines.
        """
        # Load from disk #
        try:
            df = pandas.read_csv(str(csv_path))
        # If the file is empty we can skip it #
        except pandas.errors.EmptyDataError:
            return
        # Get empty lines #
        empty_lines = df.isnull().all(1)
        # Check if there are any #
        if not any(empty_lines):
            return
        # Warn #
        msg = "The file '%s' has %i empty lines."
        raise Exception(msg % (csv_path, empty_lines.sum()))

    def reshape_events(self, debug=False):
        """Reshape the events file from the wide to the long format."""
        # The events file #
        path = self.input.paths.events
        # Optionally make a copy #
        if debug:
            path.copy(path.prefix_path + "_wide.csv")
        # Load it as a dataframe #
        wide = pandas.read_csv(str(path))
        # Reshape it #
        long = events_wide_to_long(self.country, wide)
        # Write to disk #
        long.to_csv(str(path), index=False)

    def raise_bad_timestep(self):
        """
        Raise an Exception if there are timesteps with a value below zero.
        """
        # Path to the file we want to check #
        path = str(self.input.paths.events)
        # Load from disk #
        try:
            df = pandas.read_csv(path)
        # If the file is empty we can skip it #
        except pandas.errors.EmptyDataError:
            return
        # Get negative values #
        negative_values = df["step"] < 0
        # Check if there are any #
        if not any(negative_values):
            return
        # Message #
        msg = (
            "The file '%s' has %i negative values for the timestep column."
            " This means you are attempting to apply disturbances to a"
            " year that is anterior to the inventory start year configured."
        )
        # Raise #
        raise Exception(msg % (path, negative_values.sum()))

    def copy_and_change_aidb(self):
        """Copy the AIDB and modify the disturbance matrix

        We deliberately keep both the copy and change operations in the same
        method. Because we want to be sure that the change happens only on a
        copied AIDB. Not on the reference one.
        """
        # Cases for which the default AIDB should be used
        if self.runner.silv.dist_matrix_value.use_default_aidb:
            return

        # If a disturbance matrix is defined in the yaml file
        # Check that the chosen scenario exists in disturbance_matrix_value.csv
        # Raise an error if not


        # Load the reference disturbance matrix values and the new values
        dist_matrix_table_name = "disturbance_matrix_value"
        dm = self.runner.country.aidb.db.read_df(dist_matrix_table_name)
        dm_new = self.runner.silv.dist_matrix_value.df

        # Find matching rows in the default AIDB
        ids = ["disturbance_matrix_id", "source_pool_id", "sink_pool_id"]
        df_match = dm.merge(dm_new[ids + ["proportion"]], on=ids, how="right")

        # Remove those combinations of disturbance_matrix_id and source_pool_id
        # from the table
        df_match_id = df_match.value_counts(
            ["disturbance_matrix_id", "source_pool_id"]
        ).reset_index()
        df = dm.merge(df_match_id, how="left", indicator=True)
        # Check the indicator name didn't change
        assert any(df["_merge"].str.contains("left_only"))
        selector = df["_merge"] == "left_only"
        df = df.loc[selector].copy()
        df = df[ids + ["proportion"]]

        # Add the updated disturbance matrix values and reorder
        df = pandas.concat([df, dm_new[ids + ["proportion"]]])
        df = df.sort_values(ids)
        # Drop the first index because it contains ids from both
        # the old and new data frame
        df.reset_index(drop=True, inplace=True)
        df.reset_index(inplace=True)

        # Copy the default AIDB to a temporary location. This has been added
        # because the /eos large file system Doesn't handle database writes
        # very well on JRC's BDAP computing cluster.
        combo_name = self.runner.combo.short_name
        orig_file = self.runner.country.aidb.default_aidb_path
        self.parent.log.info("AIDB %s" % orig_file)
        aidb_file_name = f"aidb_{combo_name}.db"
        with tempfile.TemporaryDirectory() as tmpdirname:
            temp_file = tmpdirname + "/" + aidb_file_name
            self.parent.log.info("Temporarily copied to %s" % temp_file)
            shutil.copy(orig_file, temp_file)
            temp_db = SQLiteDatabase(temp_file)
            # Write to the temporary AIDB
            temp_db.write_df(df, dist_matrix_table_name)
            # Change path to the modified AIDB
            self.runner.country.aidb.change_path(combo_name=combo_name)
            # Copy the temporary AIDB to the new path
            dest_file = self.runner.country.aidb.paths.aidb
            shutil.copy(temp_file, dest_file)
            self.parent.log.info("Copied to new AIDB %s" % dest_file)

        msg = "The disturbance matrix has been changed according to "
        msg += "silv/disturbance_matrix_value.csv"
        self.parent.log.info(msg)
